from datetime import datetime
import os
import unittest

from mock import call, MagicMock, patch

from yarg import newest_packages, latest_updated_packages, HTTPError
from yarg.parse import _get, Package


class GoodNewestResponse(object):
    status_code = 200
    xml = os.path.join(os.path.dirname(__file__),
                           'newest.xml')
    content = open(xml).read()


class GoodUpdatedResponse(object):
    status_code = 200
    xml = os.path.join(os.path.dirname(__file__),
                           'updated.xml')
    content = open(xml).read()


class BadResponse(object):
    status_code = 300
    reason = "Mocked"


class TestParse(unittest.TestCase):

    def setUp(self):
        self.newest = self.setup_newest()
        self.updated = self.setup_updated()

    def setup_newest(self):
        item1 = {'name': 'gobble',
                 'url': 'http://pypi.python.org/pypi/gobble',
                 'description': 'Automatic functional testing for CLI apps.',
                 'date': '09 Aug 2014 06:57:42 GMT'}
        item2 = {'name': 'flask_autorest',
                 'url': 'http://pypi.python.org/pypi/flask_autorest',
                 'description': 'auto create restful apis for database, with the help of dataset.',
                 'date': '09 Aug 2014 05:24:58 GMT'}
        item3 = {'name': 'ranrod',
                 'url': 'http://pypi.python.org/pypi/ranrod',
                 'description': 'download route53 hosted zones as local json files',
                 'date': '09 Aug 2014 05:20:21 GMT'}
        return [Package(item1), Package(item2), Package(item3)]

    def setup_updated(self):
        item1 = {'name': 'pycoin',
                 'version': '0.50',
                 'url': 'http://pypi.python.org/pypi/pycoin/0.50',
                 'description': 'Utilities for Bitcoin and altcoin addresses and transaction manipulation.',
                 'date': '09 Aug 2014 08:40:20 GMT'}
        item2 = {'name': 'django-signup',
                 'version': '0.6.0',
                 'url': 'http://pypi.python.org/pypi/django-signup/0.6.0',
                 'description': 'A user registration app for Django with support for custom user models',
                 'date': '09 Aug 2014 08:33:53 GMT'}
        item3 = {'name': 'pyADC',
                 'version': '0.1.3',
                 'url': 'http://pypi.python.org/pypi/pyADC/0.1.3',
                 'description': 'Python implementation of the ADC(S) Protocol for Direct Connect.',
                 'date': '09 Aug 2014 08:19:56 GMT'}
        return [Package(item1), Package(item2), Package(item3)]

    @patch('requests.get', return_value=BadResponse)
    def test_newest_packages_bad_get(self, get_mock):
        # Python 2.6....
        try:
            newest_packages()
        except HTTPError as e:
            self.assertEqual(300, e.status_code)
            self.assertEqual(e.status_code, e.errno)
            self.assertEqual(e.reason, e.message)

    @patch('requests.get', return_value=BadResponse)
    def test_updated_packages_bad_get(self, get_mock):
        # Python 2.6....
        try:
            latest_updated_packages()
        except HTTPError as e:
            self.assertEqual(300, e.status_code)
            self.assertEqual(e.status_code, e.errno)
            self.assertEqual(e.reason, e.message)

    @patch('requests.get', return_value=GoodNewestResponse)
    def test_newest_packages(self, get_mock):
        p = newest_packages()
        self.assertEqual(call('https://pypi.python.org/pypi?%3Aaction=packages_rss'),
                         get_mock.call_args)
        self.assertEqual(self.newest[0].name, p[0].name)
        self.assertEqual(self.newest[1].name, p[1].name)
        self.assertEqual(self.newest[2].name, p[2].name)

    @patch('requests.get', return_value=GoodNewestResponse)
    def test_newest_package(self, get_mock):
        p = newest_packages()
        self.assertEqual(call('https://pypi.python.org/pypi?%3Aaction=packages_rss'),
                         get_mock.call_args)
        self.assertEqual('gobble', p[0].name)
        self.assertEqual('http://pypi.python.org/pypi/gobble', p[0].url)
        self.assertEqual('Automatic functional testing for CLI apps.',
                         p[0].description)
        self.assertEqual(datetime.strptime('09 Aug 2014 06:57:42 GMT',
                                           "%d %b %Y %H:%M:%S %Z"),
                         p[0].date)

    @patch('requests.get', return_value=GoodNewestResponse)
    def test_newest_package_repr(self, get_mock):
        p = newest_packages()
        self.assertEqual(call('https://pypi.python.org/pypi?%3Aaction=packages_rss'),
                         get_mock.call_args)
        self.assertEqual('<Package gobble>', p[0].__repr__())

    @patch('requests.get', return_value=GoodNewestResponse)
    def test_newest_package_version(self, get_mock):
        p = newest_packages()
        self.assertEqual(call('https://pypi.python.org/pypi?%3Aaction=packages_rss'),
                         get_mock.call_args)
        self.assertEqual(None, p[0].version)

    @patch('requests.get', return_value=GoodUpdatedResponse)
    def test_updated_packages(self, get_mock):
        p = latest_updated_packages()
        self.assertEqual(call('https://pypi.python.org/pypi?%3Aaction=rss'),
                         get_mock.call_args)
        self.assertEqual(self.updated[0].name, p[0].name)
        self.assertEqual(self.updated[0].version, p[0].version)
        self.assertEqual(self.updated[1].name, p[1].name)
        self.assertEqual(self.updated[1].version, p[1].version)
        self.assertEqual(self.updated[2].name, p[2].name)
        self.assertEqual(self.updated[2].version, p[2].version)

    @patch('requests.get', return_value=GoodUpdatedResponse)
    def test_updated_package(self, get_mock):
        p = latest_updated_packages()
        self.assertEqual(call('https://pypi.python.org/pypi?%3Aaction=rss'),
                         get_mock.call_args)
        self.assertEqual('pycoin', p[0].name)
        self.assertEqual('0.50', p[0].version)
        self.assertEqual('http://pypi.python.org/pypi/pycoin/0.50', p[0].url)
        self.assertEqual('Utilities for Bitcoin and altcoin addresses and transaction manipulation.',
                         p[0].description)
        self.assertEqual(datetime.strptime('09 Aug 2014 08:40:20 GMT',
                                           "%d %b %Y %H:%M:%S %Z"),
                         p[0].date)

    @patch('requests.get', return_value=GoodUpdatedResponse)
    def test_updated_package_repr(self, get_mock):
        p = latest_updated_packages()
        self.assertEqual(call('https://pypi.python.org/pypi?%3Aaction=rss'),
                         get_mock.call_args)
        self.assertEqual('<Package pycoin>', p[0].__repr__())



# -*- coding: utf-8 -*-

import hashlib
import datetime
import math
from .contracts.store import Store
from .helpers import value
from .utils import encode


class TaggedCache(Store):
    """

    """
    def __init__(self, store, tags):
        """
        :param store: The cache store implementation
        :type store: cachy.contracts.store.Store

        :param tags: The tag set
        :type tags: cachy.tag_set.TagSet
        """
        self._store = store
        self._tags = tags

    def has(self, key):
        """
        Determine if an item exists in the cache.

        :param key: The cache key
        :type key: str

        :rtype: bool
        """
        return self.get(key) is not None

    def get(self, key, default=None):
        """
        Retrieve an item from the cache by key.

        :param key: The cache key
        :type key: str

        :param default: The default value
        :type default: mixed

        :return: The cache value
        """
        val = self._store.get(self.tagged_item_key(key))

        if val is not None:
            return val

        return value(default)

    def put(self, key, value, minutes):
        """
        Store an item in the cache for a given number of minutes.

        :param key: The cache key
        :type key: str

        :param value: The cache value
        :type value: mixed

        :param minutes: The lifetime in minutes of the cached value
        :type minutes: int or datetime
        """
        minutes = self._get_minutes(minutes)

        if minutes is not None:
            return self._store.put(self.tagged_item_key(key), value, minutes)

    def add(self, key, val, minutes):
        """
        Store an item in the cache if it does not exist.

        :param key: The cache key
        :type key: str

        :param val: The cache value
        :type val: mixed

        :param minutes: The lifetime in minutes of the cached value
        :type minutes: int|datetime

        :rtype: bool
        """
        if not self.has(key):
            self.put(key, val, minutes)

            return True

        return False

    def increment(self, key, value=1):
        """
        Increment the value of an item in the cache.

        :param key: The cache key
        :type key: str

        :param value: The increment value
        :type value: int

        :rtype: int or bool
        """
        self._store.increment(self.tagged_item_key(key), value)

    def decrement(self, key, value=1):
        """
        Decrement the value of an item in the cache.

        :param key: The cache key
        :type key: str

        :param value: The decrement value
        :type value: int

        :rtype: int or bool
        """
        self._store.decrement(self.tagged_item_key(key), value)

    def forever(self, key, value):
        """
        Store an item in the cache indefinitely.

        :param key: The cache key
        :type key: str

        :param value: The value
        :type value: mixed
        """
        self._store.forever(self.tagged_item_key(key), value)

    def forget(self, key):
        """
        Remove an item from the cache.

        :param key: The cache key
        :type key: str

        :rtype: bool
        """
        self._store.forget(self.tagged_item_key(key))

    def flush(self):
        """
        Remove all items from the cache.
        """
        self._tags.reset()

    def remember(self, key, minutes, callback):
        """
        Get an item from the cache, or store the default value.

        :param key: The cache key
        :type key: str

        :param minutes: The lifetime in minutes of the cached value
        :type minutes: int or datetime

        :param callback: The default function
        :type callback: mixed

        :rtype: mixed
        """
        # If the item exists in the cache we will just return this immediately
        # otherwise we will execute the given callback and cache the result
        # of that execution for the given number of minutes in storage.
        val = self.get(key)
        if val is not None:
            return val

        val = value(callback)

        self.put(key, val, minutes)

        return val

    def remember_forever(self, key, callback):
        """
        Get an item from the cache, or store the default value forever.

        :param key: The cache key
        :type key: str

        :param callback: The default function
        :type callback: mixed

        :rtype: mixed
        """
        # If the item exists in the cache we will just return this immediately
        # otherwise we will execute the given callback and cache the result
        # of that execution forever.
        val = self.get(key)
        if val is not None:
            return val

        val = value(callback)

        self.forever(key, val)

        return val

    def tagged_item_key(self, key):
        """
        Get a fully qualified key for a tagged item.

        :param key: The cache key
        :type key: str

        :rtype: str
        """
        return '%s:%s' % (hashlib.sha1(encode(self._tags.get_namespace())).hexdigest(), key)

    def get_prefix(self):
        """
        Get the cache key prefix.

        :rtype: str
        """
        return self._store.get_prefix()

    def _get_minutes(self, duration):
        """
        Calculate the number of minutes with the given duration.

        :param duration: The duration
        :type duration: int or datetime

        :rtype: int or None
        """
        if isinstance(duration, datetime.datetime):
            from_now = (duration - datetime.datetime.now()).total_seconds()
            from_now = math.ceil(from_now / 60)

            if from_now > 0:
                return from_now

            return

        return duration



from bleach.linkifier import (
    DEFAULT_CALLBACKS,
    Linker,
)
from bleach.sanitizer import (
    ALLOWED_ATTRIBUTES,
    ALLOWED_PROTOCOLS,
    ALLOWED_TAGS,
    Cleaner,
)


# yyyymmdd
__releasedate__ = "20220407"
# x.y.z or x.y.z.dev0 -- semver
__version__ = "5.0.0"


__all__ = ["clean", "linkify"]


def clean(
    text,
    tags=ALLOWED_TAGS,
    attributes=ALLOWED_ATTRIBUTES,
    protocols=ALLOWED_PROTOCOLS,
    strip=False,
    strip_comments=True,
    css_sanitizer=None,
):
    """Clean an HTML fragment of malicious content and return it

    This function is a security-focused function whose sole purpose is to
    remove malicious content from a string such that it can be displayed as
    content in a web page.

    This function is not designed to use to transform content to be used in
    non-web-page contexts.

    Example::

        import bleach

        better_text = bleach.clean(yucky_text)


    .. Note::

       If you're cleaning a lot of text and passing the same argument values or
       you want more configurability, consider using a
       :py:class:`bleach.sanitizer.Cleaner` instance.

    :arg str text: the text to clean

    :arg list tags: allowed list of tags; defaults to
        ``bleach.sanitizer.ALLOWED_TAGS``

    :arg dict attributes: allowed attributes; can be a callable, list or dict;
        defaults to ``bleach.sanitizer.ALLOWED_ATTRIBUTES``

    :arg list protocols: allowed list of protocols for links; defaults
        to ``bleach.sanitizer.ALLOWED_PROTOCOLS``

    :arg bool strip: whether or not to strip disallowed elements

    :arg bool strip_comments: whether or not to strip HTML comments

    :arg CSSSanitizer css_sanitizer: instance with a "sanitize_css" method for
        sanitizing style attribute values and style text; defaults to None

    :returns: cleaned text as unicode

    """
    cleaner = Cleaner(
        tags=tags,
        attributes=attributes,
        protocols=protocols,
        strip=strip,
        strip_comments=strip_comments,
        css_sanitizer=css_sanitizer,
    )
    return cleaner.clean(text)


def linkify(text, callbacks=DEFAULT_CALLBACKS, skip_tags=None, parse_email=False):
    """Convert URL-like strings in an HTML fragment to links

    This function converts strings that look like URLs, domain names and email
    addresses in text that may be an HTML fragment to links, while preserving:

    1. links already in the string
    2. urls found in attributes
    3. email addresses

    linkify does a best-effort approach and tries to recover from bad
    situations due to crazy text.

    .. Note::

       If you're linking a lot of text and passing the same argument values or
       you want more configurability, consider using a
       :py:class:`bleach.linkifier.Linker` instance.

    .. Note::

       If you have text that you want to clean and then linkify, consider using
       the :py:class:`bleach.linkifier.LinkifyFilter` as a filter in the clean
       pass. That way you're not parsing the HTML twice.

    :arg str text: the text to linkify

    :arg list callbacks: list of callbacks to run when adjusting tag attributes;
        defaults to ``bleach.linkifier.DEFAULT_CALLBACKS``

    :arg list skip_tags: list of tags that you don't want to linkify the
        contents of; for example, you could set this to ``['pre']`` to skip
        linkifying contents of ``pre`` tags

    :arg bool parse_email: whether or not to linkify email addresses

    :returns: linkified text as unicode

    """
    linker = Linker(callbacks=callbacks, skip_tags=skip_tags, parse_email=parse_email)
    return linker.linkify(text)



# -*- coding: utf-8 -*-
# Copyright 2016-2018 Julien Danjou
# Copyright 2017 Elisey Zanko
# Copyright 2016 Étienne Bersac
# Copyright 2016 Joshua Harlow
# Copyright 2013-2014 Ray Holder
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from inspect import iscoroutinefunction
except ImportError:
    iscoroutinefunction = None

# Replace a conditional import with a hard-coded None so that pip does
# not attempt to use tornado even if it is present in the environment.
# If tornado is non-None, tenacity will attempt to execute some code
# that is sensitive to the version of tornado, which could break pip
# if an old version is found.
tornado = None

import sys
import threading
import typing as t
import warnings
from abc import ABCMeta, abstractmethod
from concurrent import futures


from pip._vendor import six

from pip._vendor.tenacity import _utils

# Import all built-in retry strategies for easier usage.
from .retry import retry_base  # noqa
from .retry import retry_all  # noqa
from .retry import retry_always  # noqa
from .retry import retry_any  # noqa
from .retry import retry_if_exception  # noqa
from .retry import retry_if_exception_type  # noqa
from .retry import retry_if_not_result  # noqa
from .retry import retry_if_result  # noqa
from .retry import retry_never  # noqa
from .retry import retry_unless_exception_type  # noqa
from .retry import retry_if_exception_message  # noqa
from .retry import retry_if_not_exception_message  # noqa

# Import all nap strategies for easier usage.
from .nap import sleep  # noqa
from .nap import sleep_using_event  # noqa

# Import all built-in stop strategies for easier usage.
from .stop import stop_after_attempt  # noqa
from .stop import stop_after_delay  # noqa
from .stop import stop_all  # noqa
from .stop import stop_any  # noqa
from .stop import stop_never  # noqa
from .stop import stop_when_event_set  # noqa

# Import all built-in wait strategies for easier usage.
from .wait import wait_chain  # noqa
from .wait import wait_combine  # noqa
from .wait import wait_exponential  # noqa
from .wait import wait_fixed  # noqa
from .wait import wait_incrementing  # noqa
from .wait import wait_none  # noqa
from .wait import wait_random  # noqa
from .wait import wait_random_exponential  # noqa
from .wait import wait_random_exponential as wait_full_jitter  # noqa

# Import all built-in before strategies for easier usage.
from .before import before_log  # noqa
from .before import before_nothing  # noqa

# Import all built-in after strategies for easier usage.
from .after import after_log  # noqa
from .after import after_nothing  # noqa

# Import all built-in after strategies for easier usage.
from .before_sleep import before_sleep_log  # noqa
from .before_sleep import before_sleep_nothing  # noqa


WrappedFn = t.TypeVar("WrappedFn", bound=t.Callable)


@t.overload
def retry(fn):
    # type: (WrappedFn) -> WrappedFn
    """Type signature for @retry as a raw decorator."""
    pass


@t.overload
def retry(*dargs, **dkw):  # noqa
    # type: (...) -> t.Callable[[WrappedFn], WrappedFn]
    """Type signature for the @retry() decorator constructor."""
    pass


def retry(*dargs, **dkw):  # noqa
    """Wrap a function with a new `Retrying` object.

    :param dargs: positional arguments passed to Retrying object
    :param dkw: keyword arguments passed to the Retrying object
    """
    # support both @retry and @retry() as valid syntax
    if len(dargs) == 1 and callable(dargs[0]):
        return retry()(dargs[0])
    else:

        def wrap(f):
            if isinstance(f, retry_base):
                warnings.warn(
                    (
                        "Got retry_base instance ({cls}) as callable argument, "
                        + "this will probably hang indefinitely (did you mean "
                        + "retry={cls}(...)?)"
                    ).format(cls=f.__class__.__name__)
                )
            if iscoroutinefunction is not None and iscoroutinefunction(f):
                r = AsyncRetrying(*dargs, **dkw)
            elif (
                tornado
                and hasattr(tornado.gen, "is_coroutine_function")
                and tornado.gen.is_coroutine_function(f)
            ):
                r = TornadoRetrying(*dargs, **dkw)
            else:
                r = Retrying(*dargs, **dkw)

            return r.wraps(f)

        return wrap


class TryAgain(Exception):
    """Always retry the executed function when raised."""


NO_RESULT = object()


class DoAttempt(object):
    pass


class DoSleep(float):
    pass


class BaseAction(object):
    """Base class for representing actions to take by retry object.

    Concrete implementations must define:
    - __init__: to initialize all necessary fields
    - REPR_ATTRS: class variable specifying attributes to include in repr(self)
    - NAME: for identification in retry object methods and callbacks
    """

    REPR_FIELDS = ()
    NAME = None

    def __repr__(self):
        state_str = ", ".join(
            "%s=%r" % (field, getattr(self, field)) for field in self.REPR_FIELDS
        )
        return "%s(%s)" % (type(self).__name__, state_str)

    def __str__(self):
        return repr(self)


class RetryAction(BaseAction):
    REPR_FIELDS = ("sleep",)
    NAME = "retry"

    def __init__(self, sleep):
        self.sleep = float(sleep)


_unset = object()


def _first_set(first, second):
    return second if first is _unset else first


class RetryError(Exception):
    """Encapsulates the last attempt instance right before giving up."""

    def __init__(self, last_attempt):
        self.last_attempt = last_attempt
        super(RetryError, self).__init__(last_attempt)

    def reraise(self):
        if self.last_attempt.failed:
            raise self.last_attempt.result()
        raise self

    def __str__(self):
        return "{0}[{1}]".format(self.__class__.__name__, self.last_attempt)


class AttemptManager(object):
    """Manage attempt context."""

    def __init__(self, retry_state):
        self.retry_state = retry_state

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if isinstance(exc_value, BaseException):
            self.retry_state.set_exception((exc_type, exc_value, traceback))
            return True  # Swallow exception.
        else:
            # We don't have the result, actually.
            self.retry_state.set_result(None)


class BaseRetrying(object):
    __metaclass__ = ABCMeta

    def __init__(
        self,
        sleep=sleep,
        stop=stop_never,
        wait=wait_none(),
        retry=retry_if_exception_type(),
        before=before_nothing,
        after=after_nothing,
        before_sleep=None,
        reraise=False,
        retry_error_cls=RetryError,
        retry_error_callback=None,
    ):
        self.sleep = sleep
        self.stop = stop
        self.wait = wait
        self.retry = retry
        self.before = before
        self.after = after
        self.before_sleep = before_sleep
        self.reraise = reraise
        self._local = threading.local()
        self.retry_error_cls = retry_error_cls
        self.retry_error_callback = retry_error_callback

        # This attribute was moved to RetryCallState and is deprecated on
        # Retrying objects but kept for backward compatibility.
        self.fn = None

    def copy(
        self,
        sleep=_unset,
        stop=_unset,
        wait=_unset,
        retry=_unset,
        before=_unset,
        after=_unset,
        before_sleep=_unset,
        reraise=_unset,
        retry_error_cls=_unset,
        retry_error_callback=_unset,
    ):
        """Copy this object with some parameters changed if needed."""
        return self.__class__(
            sleep=_first_set(sleep, self.sleep),
            stop=_first_set(stop, self.stop),
            wait=_first_set(wait, self.wait),
            retry=_first_set(retry, self.retry),
            before=_first_set(before, self.before),
            after=_first_set(after, self.after),
            before_sleep=_first_set(before_sleep, self.before_sleep),
            reraise=_first_set(reraise, self.reraise),
            retry_error_cls=_first_set(retry_error_cls, self.retry_error_cls),
            retry_error_callback=_first_set(
                retry_error_callback, self.retry_error_callback
            ),
        )

    def __repr__(self):
        attrs = dict(
            _utils.visible_attrs(self, attrs={"me": id(self)}),
            __class__=self.__class__.__name__,
        )
        return (
            "<%(__class__)s object at 0x%(me)x (stop=%(stop)s, "
            "wait=%(wait)s, sleep=%(sleep)s, retry=%(retry)s, "
            "before=%(before)s, after=%(after)s)>"
        ) % (attrs)

    @property
    def statistics(self):
        """Return a dictionary of runtime statistics.

        This dictionary will be empty when the controller has never been
        ran. When it is running or has ran previously it should have (but
        may not) have useful and/or informational keys and values when
        running is underway and/or completed.

        .. warning:: The keys in this dictionary **should** be some what
                     stable (not changing), but there existence **may**
                     change between major releases as new statistics are
                     gathered or removed so before accessing keys ensure that
                     they actually exist and handle when they do not.

        .. note:: The values in this dictionary are local to the thread
                  running call (so if multiple threads share the same retrying
                  object - either directly or indirectly) they will each have
                  there own view of statistics they have collected (in the
                  future we may provide a way to aggregate the various
                  statistics from each thread).
        """
        try:
            return self._local.statistics
        except AttributeError:
            self._local.statistics = {}
            return self._local.statistics

    def wraps(self, f):
        """Wrap a function for retrying.

        :param f: A function to wraps for retrying.
        """

        @_utils.wraps(f)
        def wrapped_f(*args, **kw):
            return self(f, *args, **kw)

        def retry_with(*args, **kwargs):
            return self.copy(*args, **kwargs).wraps(f)

        wrapped_f.retry = self
        wrapped_f.retry_with = retry_with

        return wrapped_f

    def begin(self, fn):
        self.statistics.clear()
        self.statistics["start_time"] = _utils.now()
        self.statistics["attempt_number"] = 1
        self.statistics["idle_for"] = 0
        self.fn = fn

    def iter(self, retry_state):  # noqa
        fut = retry_state.outcome
        if fut is None:
            if self.before is not None:
                self.before(retry_state)
            return DoAttempt()

        is_explicit_retry = retry_state.outcome.failed and isinstance(
            retry_state.outcome.exception(), TryAgain
        )
        if not (is_explicit_retry or self.retry(retry_state=retry_state)):
            return fut.result()

        if self.after is not None:
            self.after(retry_state=retry_state)

        self.statistics["delay_since_first_attempt"] = retry_state.seconds_since_start
        if self.stop(retry_state=retry_state):
            if self.retry_error_callback:
                return self.retry_error_callback(retry_state=retry_state)
            retry_exc = self.retry_error_cls(fut)
            if self.reraise:
                raise retry_exc.reraise()
            six.raise_from(retry_exc, fut.exception())

        if self.wait:
            sleep = self.wait(retry_state=retry_state)
        else:
            sleep = 0.0
        retry_state.next_action = RetryAction(sleep)
        retry_state.idle_for += sleep
        self.statistics["idle_for"] += sleep
        self.statistics["attempt_number"] += 1

        if self.before_sleep is not None:
            self.before_sleep(retry_state=retry_state)

        return DoSleep(sleep)

    def __iter__(self):
        self.begin(None)

        retry_state = RetryCallState(self, fn=None, args=(), kwargs={})
        while True:
            do = self.iter(retry_state=retry_state)
            if isinstance(do, DoAttempt):
                yield AttemptManager(retry_state=retry_state)
            elif isinstance(do, DoSleep):
                retry_state.prepare_for_next_attempt()
                self.sleep(do)
            else:
                break

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass

    def call(self, *args, **kwargs):
        """Use ``__call__`` instead because this method is deprecated."""
        warnings.warn(
            "'call()' method is deprecated. " + "Use '__call__()' instead",
            DeprecationWarning,
        )
        return self.__call__(*args, **kwargs)


class Retrying(BaseRetrying):
    """Retrying controller."""

    def __call__(self, fn, *args, **kwargs):
        self.begin(fn)

        retry_state = RetryCallState(retry_object=self, fn=fn, args=args, kwargs=kwargs)
        while True:
            do = self.iter(retry_state=retry_state)
            if isinstance(do, DoAttempt):
                try:
                    result = fn(*args, **kwargs)
                except BaseException:  # noqa: B902
                    retry_state.set_exception(sys.exc_info())
                else:
                    retry_state.set_result(result)
            elif isinstance(do, DoSleep):
                retry_state.prepare_for_next_attempt()
                self.sleep(do)
            else:
                return do


class Future(futures.Future):
    """Encapsulates a (future or past) attempted call to a target function."""

    def __init__(self, attempt_number):
        super(Future, self).__init__()
        self.attempt_number = attempt_number

    @property
    def failed(self):
        """Return whether a exception is being held in this future."""
        return self.exception() is not None

    @classmethod
    def construct(cls, attempt_number, value, has_exception):
        """Construct a new Future object."""
        fut = cls(attempt_number)
        if has_exception:
            fut.set_exception(value)
        else:
            fut.set_result(value)
        return fut


class RetryCallState(object):
    """State related to a single call wrapped with Retrying."""

    def __init__(self, retry_object, fn, args, kwargs):
        #: Retry call start timestamp
        self.start_time = _utils.now()
        #: Retry manager object
        self.retry_object = retry_object
        #: Function wrapped by this retry call
        self.fn = fn
        #: Arguments of the function wrapped by this retry call
        self.args = args
        #: Keyword arguments of the function wrapped by this retry call
        self.kwargs = kwargs

        #: The number of the current attempt
        self.attempt_number = 1
        #: Last outcome (result or exception) produced by the function
        self.outcome = None
        #: Timestamp of the last outcome
        self.outcome_timestamp = None
        #: Time spent sleeping in retries
        self.idle_for = 0
        #: Next action as decided by the retry manager
        self.next_action = None

    @property
    def seconds_since_start(self):
        if self.outcome_timestamp is None:
            return None
        return self.outcome_timestamp - self.start_time

    def prepare_for_next_attempt(self):
        self.outcome = None
        self.outcome_timestamp = None
        self.attempt_number += 1
        self.next_action = None

    def set_result(self, val):
        ts = _utils.now()
        fut = Future(self.attempt_number)
        fut.set_result(val)
        self.outcome, self.outcome_timestamp = fut, ts

    def set_exception(self, exc_info):
        ts = _utils.now()
        fut = Future(self.attempt_number)
        _utils.capture(fut, exc_info)
        self.outcome, self.outcome_timestamp = fut, ts


if iscoroutinefunction:
    from pip._vendor.tenacity._asyncio import AsyncRetrying

if tornado:
    from pip._vendor.tenacity.tornadoweb import TornadoRetrying



from typing import Any
from typing import Optional

from clikit.utils.string import parse_boolean
from clikit.utils.string import parse_float
from clikit.utils.string import parse_int
from clikit.utils.string import parse_string

from .abstract_option import AbstractOption


class Option(AbstractOption):
    """
    An input option
    """

    NO_VALUE = 4
    REQUIRED_VALUE = 8
    OPTIONAL_VALUE = 16
    MULTI_VALUED = 32

    STRING = 128
    BOOLEAN = 256
    INTEGER = 512
    FLOAT = 1024
    NULLABLE = 2048

    def __init__(
        self,
        long_name,
        short_name=None,
        flags=0,
        description=None,
        default=None,
        value_name="...",
    ):  # type: (str, Optional[str], int, Optional[str], Any, str) -> None
        self._validate_flags(flags)

        super(Option, self).__init__(long_name, short_name, flags, description)

        self._value_name = value_name

        if self.is_multi_valued():
            self._default = []
        else:
            self._default = None

        if self.accepts_value() or default is not None:
            self.set_default(default)

    @property
    def default(self):  # type: () -> Any
        return self._default

    @property
    def value_name(self):  # type: () -> str
        return self._value_name

    def accepts_value(self):  # type: () -> bool
        return not bool(self.NO_VALUE & self._flags)

    def parse(self, value):  # type: (Any) -> Any
        nullable = bool(self._flags & self.NULLABLE)

        if self._flags & self.BOOLEAN:
            return parse_boolean(value, nullable)
        elif self._flags & self.INTEGER:
            return parse_int(value, nullable)
        elif self._flags & self.FLOAT:
            return parse_float(value, nullable)

        return parse_string(value, nullable)

    def is_value_required(self):  # type: () -> bool
        return bool(self.REQUIRED_VALUE & self._flags)

    def is_value_optional(self):  # type: () -> bool
        return bool(self.OPTIONAL_VALUE & self._flags)

    def is_multi_valued(self):  # type: () -> bool
        return bool(self.MULTI_VALUED & self._flags)

    def set_default(self, default=None):  # type: (Any) -> None
        if not self.accepts_value():
            raise ValueError(
                "Cannot set a default value when using the flag VALUE_NONE."
            )

        if self.is_multi_valued():
            if default is None:
                default = []
            elif not isinstance(default, list):
                raise ValueError(
                    "The default value of a multi-valued option must be a list. "
                    "Got: {}".format(type(default))
                )

        self._default = default

    def _validate_flags(self, flags):  # type: (int) -> None
        super(Option, self)._validate_flags(flags)

        if flags & self.NO_VALUE:
            if flags & self.REQUIRED_VALUE:
                raise ValueError(
                    "The option flags VALUE_NONE and VALUE_REQUIRED cannot be combined."
                )

            if flags & self.OPTIONAL_VALUE:
                raise ValueError(
                    "The option flags VALUE_NONE and VALUE_OPTIONAL cannot be combined."
                )

            if flags & self.MULTI_VALUED:
                raise ValueError(
                    "The option flags VALUE_NONE and MULTI_VALUED cannot be combined."
                )

        if flags & self.OPTIONAL_VALUE and flags & self.MULTI_VALUED:
            raise ValueError(
                "The option flags VALUE_OPTIONAL and MULTI_VALUED cannot be combined."
            )

        if flags & self.STRING:
            if flags & self.BOOLEAN:
                raise ValueError(
                    "The option flags STRING and BOOLEAN cannot be combined."
                )

            if flags & self.INTEGER:
                raise ValueError(
                    "The option flags STRING and INTEGER cannot be combined."
                )

            if flags & self.FLOAT:
                raise ValueError(
                    "The option flags STRING and FLOAT cannot be combined."
                )
        elif flags & self.BOOLEAN:
            if flags & self.INTEGER:
                raise ValueError(
                    "The option flags BOOLEAN and INTEGER cannot be combined."
                )

            if flags & self.FLOAT:
                raise ValueError(
                    "The option flags BOOLEAN and FLOAT cannot be combined."
                )
        elif flags & self.INTEGER:
            if flags & self.FLOAT:
                raise ValueError(
                    "The option flags INTEGER and FLOAT cannot be combined."
                )

    def _add_default_flags(self, flags):  # type: (int) -> int
        flags = super(Option, self)._add_default_flags(flags)

        if not flags & (
            self.NO_VALUE
            | self.REQUIRED_VALUE
            | self.OPTIONAL_VALUE
            | self.MULTI_VALUED
        ):
            flags |= self.NO_VALUE

        if not flags & (self.STRING | self.BOOLEAN | self.INTEGER | self.FLOAT):
            flags |= self.STRING

        if flags & self.MULTI_VALUED and not flags & self.REQUIRED_VALUE:
            flags |= self.REQUIRED_VALUE

        return flags



import math


_TIME_FORMATS = [
    (0, "< 1 sec"),
    (2, "1 sec"),
    (59, "secs", 1),
    (60, "1 min"),
    (3600, "mins", 60),
    (5400, "1 hr"),
    (86400, "hrs", 3600),
    (129600, "1 day"),
    (604800, "days", 86400),
]


def format_time(secs):  # type: (int) -> str
    for fmt in _TIME_FORMATS:
        if secs > fmt[0]:
            continue

        if len(fmt) == 2:
            return fmt[1]

        return "{} {}".format(math.ceil(secs / fmt[2]), fmt[1])



"""
Tests for dict duplicate keys Pyflakes behavior.
"""

from sys import version_info

from pyflakes import messages as m
from pyflakes.test.harness import TestCase, skipIf


class Test(TestCase):

    def test_duplicate_keys(self):
        self.flakes(
            "{'yes': 1, 'yes': 2}",
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    @skipIf(version_info < (3,),
            "bytes and strings with same 'value' are not equal in python3")
    def test_duplicate_keys_bytes_vs_unicode_py3(self):
        self.flakes("{b'a': 1, u'a': 2}")

    @skipIf(version_info < (3,),
            "bytes and strings with same 'value' are not equal in python3")
    def test_duplicate_values_bytes_vs_unicode_py3(self):
        self.flakes(
            "{1: b'a', 1: u'a'}",
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    @skipIf(version_info >= (3,),
            "bytes and strings with same 'value' are equal in python2")
    def test_duplicate_keys_bytes_vs_unicode_py2(self):
        self.flakes(
            "{b'a': 1, u'a': 2}",
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    @skipIf(version_info >= (3,),
            "bytes and strings with same 'value' are equal in python2")
    def test_duplicate_values_bytes_vs_unicode_py2(self):
        self.flakes("{1: b'a', 1: u'a'}")

    def test_multiple_duplicate_keys(self):
        self.flakes(
            "{'yes': 1, 'yes': 2, 'no': 2, 'no': 3}",
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    def test_duplicate_keys_in_function(self):
        self.flakes(
            '''
            def f(thing):
                pass
            f({'yes': 1, 'yes': 2})
            ''',
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    def test_duplicate_keys_in_lambda(self):
        self.flakes(
            "lambda x: {(0,1): 1, (0,1): 2}",
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    def test_duplicate_keys_tuples(self):
        self.flakes(
            "{(0,1): 1, (0,1): 2}",
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    def test_duplicate_keys_tuples_int_and_float(self):
        self.flakes(
            "{(0,1): 1, (0,1.0): 2}",
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    def test_duplicate_keys_ints(self):
        self.flakes(
            "{1: 1, 1: 2}",
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    def test_duplicate_keys_bools(self):
        self.flakes(
            "{True: 1, True: 2}",
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    def test_duplicate_keys_bools_false(self):
        # Needed to ensure 2.x correctly coerces these from variables
        self.flakes(
            "{False: 1, False: 2}",
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    def test_duplicate_keys_none(self):
        self.flakes(
            "{None: 1, None: 2}",
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    def test_duplicate_variable_keys(self):
        self.flakes(
            '''
            a = 1
            {a: 1, a: 2}
            ''',
            m.MultiValueRepeatedKeyVariable,
            m.MultiValueRepeatedKeyVariable,
        )

    def test_duplicate_variable_values(self):
        self.flakes(
            '''
            a = 1
            b = 2
            {1: a, 1: b}
            ''',
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    def test_duplicate_variable_values_same_value(self):
        # Current behaviour is not to look up variable values. This is to
        # confirm that.
        self.flakes(
            '''
            a = 1
            b = 1
            {1: a, 1: b}
            ''',
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    def test_duplicate_key_float_and_int(self):
        """
        These do look like different values, but when it comes to their use as
        keys, they compare as equal and so are actually duplicates.
        The literal dict {1: 1, 1.0: 1} actually becomes {1.0: 1}.
        """
        self.flakes(
            '''
            {1: 1, 1.0: 2}
            ''',
            m.MultiValueRepeatedKeyLiteral,
            m.MultiValueRepeatedKeyLiteral,
        )

    def test_no_duplicate_key_error_same_value(self):
        self.flakes('''
        {'yes': 1, 'yes': 1}
        ''')

    def test_no_duplicate_key_errors(self):
        self.flakes('''
        {'yes': 1, 'no': 2}
        ''')

    def test_no_duplicate_keys_tuples_same_first_element(self):
        self.flakes("{(0,1): 1, (0,2): 1}")

    def test_no_duplicate_key_errors_func_call(self):
        self.flakes('''
        def test(thing):
            pass
        test({True: 1, None: 2, False: 1})
        ''')

    def test_no_duplicate_key_errors_bool_or_none(self):
        self.flakes("{True: 1, None: 2, False: 1}")

    def test_no_duplicate_key_errors_ints(self):
        self.flakes('''
        {1: 1, 2: 1}
        ''')

    def test_no_duplicate_key_errors_vars(self):
        self.flakes('''
        test = 'yes'
        rest = 'yes'
        {test: 1, rest: 2}
        ''')

    def test_no_duplicate_key_errors_tuples(self):
        self.flakes('''
        {(0,1): 1, (0,2): 1}
        ''')

    def test_no_duplicate_key_errors_instance_attributes(self):
        self.flakes('''
        class Test():
            pass
        f = Test()
        f.a = 1
        {f.a: 1, f.a: 1}
        ''')



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.


import sys
import typing

from cryptography import utils
from cryptography.exceptions import (
    AlreadyFinalized,
    InvalidKey,
    UnsupportedAlgorithm,
    _Reasons,
)
from cryptography.hazmat.backends import _get_backend
from cryptography.hazmat.backends.interfaces import Backend, ScryptBackend
from cryptography.hazmat.primitives import constant_time
from cryptography.hazmat.primitives.kdf import KeyDerivationFunction


# This is used by the scrypt tests to skip tests that require more memory
# than the MEM_LIMIT
_MEM_LIMIT = sys.maxsize // 2


class Scrypt(KeyDerivationFunction):
    def __init__(
        self,
        salt: bytes,
        length: int,
        n: int,
        r: int,
        p: int,
        backend: typing.Optional[Backend] = None,
    ):
        backend = _get_backend(backend)
        if not isinstance(backend, ScryptBackend):
            raise UnsupportedAlgorithm(
                "Backend object does not implement ScryptBackend.",
                _Reasons.BACKEND_MISSING_INTERFACE,
            )

        self._length = length
        utils._check_bytes("salt", salt)
        if n < 2 or (n & (n - 1)) != 0:
            raise ValueError("n must be greater than 1 and be a power of 2.")

        if r < 1:
            raise ValueError("r must be greater than or equal to 1.")

        if p < 1:
            raise ValueError("p must be greater than or equal to 1.")

        self._used = False
        self._salt = salt
        self._n = n
        self._r = r
        self._p = p
        self._backend = backend

    def derive(self, key_material: bytes) -> bytes:
        if self._used:
            raise AlreadyFinalized("Scrypt instances can only be used once.")
        self._used = True

        utils._check_byteslike("key_material", key_material)
        return self._backend.derive_scrypt(
            key_material, self._salt, self._length, self._n, self._r, self._p
        )

    def verify(self, key_material: bytes, expected_key: bytes) -> None:
        derived_key = self.derive(key_material)
        if not constant_time.bytes_eq(derived_key, expected_key):
            raise InvalidKey("Keys do not match.")



from rope.base import pynames, evaluate, exceptions, worder
from rope.refactor.rename import Rename


class LocalToField(object):

    def __init__(self, project, resource, offset):
        self.project = project
        self.resource = resource
        self.offset = offset

    def get_changes(self):
        name = worder.get_name_at(self.resource, self.offset)
        this_pymodule = self.project.get_pymodule(self.resource)
        pyname = evaluate.eval_location(this_pymodule, self.offset)
        if not self._is_a_method_local(pyname):
            raise exceptions.RefactoringError(
                'Convert local variable to field should be performed on \n'
                'a local variable of a method.')

        pymodule, lineno = pyname.get_definition_location()
        function_scope = pymodule.get_scope().get_inner_scope_for_line(lineno)
        # Not checking redefinition
        #self._check_redefinition(name, function_scope)

        new_name = self._get_field_name(function_scope.pyobject, name)
        changes = Rename(self.project, self.resource, self.offset).\
            get_changes(new_name, resources=[self.resource])
        return changes

    def _check_redefinition(self, name, function_scope):
        class_scope = function_scope.parent
        if name in class_scope.pyobject:
            raise exceptions.RefactoringError(
                'The field %s already exists' % name)

    def _get_field_name(self, pyfunction, name):
        self_name = pyfunction.get_param_names()[0]
        new_name = self_name + '.' + name
        return new_name

    def _is_a_method_local(self, pyname):
        pymodule, lineno = pyname.get_definition_location()
        holding_scope = pymodule.get_scope().get_inner_scope_for_line(lineno)
        parent = holding_scope.parent
        return isinstance(pyname, pynames.AssignedName) and \
            pyname in holding_scope.get_names().values() and \
            holding_scope.get_kind() == 'Function' and \
            parent is not None and parent.get_kind() == 'Class'



import itertools
import json
import os
import re

from jsonschema.compat import MutableMapping, str_types, urlsplit


class URIDict(MutableMapping):
    """
    Dictionary which uses normalized URIs as keys.
    """

    def normalize(self, uri):
        return urlsplit(uri).geturl()

    def __init__(self, *args, **kwargs):
        self.store = dict()
        self.store.update(*args, **kwargs)

    def __getitem__(self, uri):
        return self.store[self.normalize(uri)]

    def __setitem__(self, uri, value):
        self.store[self.normalize(uri)] = value

    def __delitem__(self, uri):
        del self.store[self.normalize(uri)]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __repr__(self):
        return repr(self.store)


class Unset(object):
    """
    An as-of-yet unset attribute or unprovided default parameter.
    """

    def __repr__(self):
        return "<unset>"


def load_schema(name):
    """
    Load a schema from ./schemas/``name``.json and return it.
    """
    with open(
        os.path.join(os.path.dirname(__file__), "schemas", "{0}.json".format(name))
    ) as f:
        data = f.read()

    return json.loads(data)


def indent(string, times=1):
    """
    A dumb version of `textwrap.indent` from Python 3.3.
    """

    return "\n".join(" " * (4 * times) + line for line in string.splitlines())


def format_as_index(indices):
    """
    Construct a single string containing indexing operations for the indices.

    For example, [1, 2, "foo"] -> [1][2]["foo"]

    Arguments:

        indices (sequence):

            The indices to format.
    """

    if not indices:
        return ""
    return "[%s]" % "][".join(repr(index) for index in indices)


def find_additional_properties(instance, schema):
    """
    Return the set of additional properties for the given ``instance``.

    Weeds out properties that should have been validated by ``properties`` and
    / or ``patternProperties``.

    Assumes ``instance`` is dict-like already.
    """

    properties = schema.get("properties", {})
    patterns = "|".join(schema.get("patternProperties", {}))
    for property in instance:
        if property not in properties:
            if patterns and re.search(patterns, property):
                continue
            yield property


def extras_msg(extras):
    """
    Create an error message for extra items or properties.
    """

    if len(extras) == 1:
        verb = "was"
    else:
        verb = "were"
    return ", ".join(repr(extra) for extra in extras), verb


def types_msg(instance, types):
    """
    Create an error message for a failure to match the given types.

    If the ``instance`` is an object and contains a ``name`` property, it will
    be considered to be a description of that object and used as its type.

    Otherwise the message is simply the reprs of the given ``types``.
    """

    reprs = []
    for type in types:
        try:
            reprs.append(repr(type["name"]))
        except Exception:
            reprs.append(repr(type))
    return "%r is not of type %s" % (instance, ", ".join(reprs))


def flatten(suitable_for_isinstance):
    """
    isinstance() can accept a bunch of really annoying different types:
        * a single type
        * a tuple of types
        * an arbitrary nested tree of tuples

    Return a flattened tuple of the given argument.
    """

    types = set()

    if not isinstance(suitable_for_isinstance, tuple):
        suitable_for_isinstance = (suitable_for_isinstance,)
    for thing in suitable_for_isinstance:
        if isinstance(thing, tuple):
            types.update(flatten(thing))
        else:
            types.add(thing)
    return tuple(types)


def ensure_list(thing):
    """
    Wrap ``thing`` in a list if it's a single str.

    Otherwise, return it unchanged.
    """

    if isinstance(thing, str_types):
        return [thing]
    return thing


def equal(one, two):
    """
    Check if two things are equal, but evade booleans and ints being equal.
    """
    return unbool(one) == unbool(two)


def unbool(element, true=object(), false=object()):
    """
    A hack to make True and 1 and False and 0 unique for ``uniq``.
    """

    if element is True:
        return true
    elif element is False:
        return false
    return element


def uniq(container):
    """
    Check if all of a container's elements are unique.

    Successively tries first to rely that the elements are hashable, then
    falls back on them being sortable, and finally falls back on brute
    force.
    """

    try:
        return len(set(unbool(i) for i in container)) == len(container)
    except TypeError:
        try:
            sort = sorted(unbool(i) for i in container)
            sliced = itertools.islice(sort, 1, None)
            for i, j in zip(sort, sliced):
                if i == j:
                    return False
        except (NotImplementedError, TypeError):
            seen = []
            for e in container:
                e = unbool(e)
                if e in seen:
                    return False
                seen.append(e)
    return True



# -*- coding: utf-8 -*-

import time
import math
from ..contracts.taggable_store import TaggableStore


class DictStore(TaggableStore):
    """
    A cache store using a dictionary as its backend.
    """

    def __init__(self):
        self._storage = {}

    def get(self, key):
        """
        Retrieve an item from the cache by key.

        :param key: The cache key
        :type key: str

        :return: The cache value
        """
        return self._get_payload(key)[0]

    def _get_payload(self, key):
        """
        Retrieve an item and expiry time from the cache by key.

        :param key: The cache key
        :type key: str

        :rtype: dict
        """
        payload = self._storage.get(key)

        # If the key does not exist, we return nothing
        if not payload:
            return (None, None)

        expire = payload[0]

        # If the current time is greater than expiration timestamps we will delete
        # the entry
        if round(time.time()) >= expire:
            self.forget(key)

            return (None, None)

        data = payload[1]

        # Next, we'll extract the number of minutes that are remaining for a cache
        # so that we can properly retain the time for things like the increment
        # operation that may be performed on the cache. We'll round this out.
        time_ = math.ceil((expire - round(time.time())) / 60.)

        return (data, time_)

    def put(self, key, value, minutes):
        """
        Store an item in the cache for a given number of minutes.

        :param key: The cache key
        :type key: str

        :param value: The cache value
        :type value: mixed

        :param minutes: The lifetime in minutes of the cached value
        :type minutes: int
        """
        self._storage[key] = (self._expiration(minutes), value)

    def increment(self, key, value=1):
        """
        Increment the value of an item in the cache.

        :param key: The cache key
        :type key: str

        :param value: The increment value
        :type value: int

        :rtype: int or bool
        """
        data, time_ = self._get_payload(key)

        integer = int(data) + value

        self.put(key, integer, int(time_))

        return integer

    def decrement(self, key, value=1):
        """
        Decrement the value of an item in the cache.

        :param key: The cache key
        :type key: str

        :param value: The decrement value
        :type value: int

        :rtype: int or bool
        """
        return self.increment(key, value * -1)

    def forever(self, key, value):
        """
        Store an item in the cache indefinitely.

        :param key: The cache key
        :type key: str

        :param value: The increment value
        :type value: int
        """
        self.put(key, value, 0)

    def forget(self, key):
        """
        Remove an item from the cache.

        :param key: The cache key
        :type key: str

        :rtype: bool
        """
        if key in self._storage:
            del self._storage[key]

            return True

        return False

    def flush(self):
        """
        Remove all items from the cache.
        """
        self._storage = {}

    def _expiration(self, minutes):
        """
        Get the expiration time based on the given minutes.

        :param minutes: The minutes
        :type minutes: int

        :rtype: int
        """
        if minutes == 0:
            return 9999999999

        return round(time.time()) + (minutes * 60)

    def get_prefix(self):
        """
        Get the cache key prefix.

        :rtype: str
        """
        return ''



"""Find occurrences of a name in a project.

This module consists of a `Finder` that finds all occurrences of a name
in a project. The `Finder.find_occurrences()` method is a generator that
yields `Occurrence` instances for each occurrence of the name. To create
a `Finder` object, use the `create_finder()` function:

    finder = occurrences.create_finder(project, 'foo', pyname)
    for occurrence in finder.find_occurrences():
        pass

It's possible to filter the occurrences. They can be specified when
calling the `create_finder()` function.

  * `only_calls`: If True, return only those instances where the name is
    a function that's being called.

  * `imports`: If False, don't return instances that are in import
    statements.

  * `unsure`: If a prediate function, return instances where we don't
    know what the name references. It also filters based on the
    predicate function.

  * `docs`: If True, it will search for occurrences in regions normally
    ignored. E.g., strings and comments.

  * `in_hierarchy`: If True, it will find occurrences if the name is in
    the class's hierarchy.

  * `instance`: Used only when you want implicit interfaces to be
    considered.

  * `keywords`: If False, don't return instances that are the names of keyword
    arguments
"""

import ast
import re

from rope.base import codeanalyze
from rope.base import evaluate
from rope.base import exceptions
from rope.base import pynames
from rope.base import pyobjects
from rope.base import utils
from rope.base import worder


class Finder(object):
    """For finding occurrences of a name

    The constructor takes a `filters` argument.  It should be a list
    of functions that take a single argument.  For each possible
    occurrence, these functions are called in order with the an
    instance of `Occurrence`:

      * If it returns `None` other filters are tried.
      * If it returns `True`, the occurrence will be a match.
      * If it returns `False`, the occurrence will be skipped.
      * If all of the filters return `None`, it is skipped also.

    """

    def __init__(self, project, name, filters=[lambda o: True], docs=False):
        self.project = project
        self.name = name
        self.docs = docs
        self.filters = filters
        self._textual_finder = _TextualFinder(name, docs=docs)

    def find_occurrences(self, resource=None, pymodule=None):
        """Generate `Occurrence` instances"""
        tools = _OccurrenceToolsCreator(self.project, resource=resource,
                                        pymodule=pymodule, docs=self.docs)
        for offset in self._textual_finder.find_offsets(tools.source_code):
            occurrence = Occurrence(tools, offset)
            for filter in self.filters:
                result = filter(occurrence)
                if result is None:
                    continue
                if result:
                    yield occurrence
                break


def create_finder(project, name, pyname, only_calls=False, imports=True,
                  unsure=None, docs=False, instance=None, in_hierarchy=False,
                  keywords=True):
    """A factory for `Finder`

    Based on the arguments it creates a list of filters.  `instance`
    argument is needed only when you want implicit interfaces to be
    considered.

    """
    pynames_ = set([pyname])
    filters = []
    if only_calls:
        filters.append(CallsFilter())
    if not imports:
        filters.append(NoImportsFilter())
    if not keywords:
        filters.append(NoKeywordsFilter())
    if isinstance(instance, pynames.ParameterName):
        for pyobject in instance.get_objects():
            try:
                pynames_.add(pyobject[name])
            except exceptions.AttributeNotFoundError:
                pass
    for pyname in pynames_:
        filters.append(PyNameFilter(pyname))
        if in_hierarchy:
            filters.append(InHierarchyFilter(pyname))
    if unsure:
        filters.append(UnsureFilter(unsure))
    return Finder(project, name, filters=filters, docs=docs)


class Occurrence(object):

    def __init__(self, tools, offset):
        self.tools = tools
        self.offset = offset
        self.resource = tools.resource

    @utils.saveit
    def get_word_range(self):
        return self.tools.word_finder.get_word_range(self.offset)

    @utils.saveit
    def get_primary_range(self):
        return self.tools.word_finder.get_primary_range(self.offset)

    @utils.saveit
    def get_pyname(self):
        try:
            return self.tools.name_finder.get_pyname_at(self.offset)
        except exceptions.BadIdentifierError:
            pass

    @utils.saveit
    def get_primary_and_pyname(self):
        try:
            return self.tools.name_finder.get_primary_and_pyname_at(
                self.offset)
        except exceptions.BadIdentifierError:
            pass

    @utils.saveit
    def is_in_import_statement(self):
        return (self.tools.word_finder.is_from_statement(self.offset) or
                self.tools.word_finder.is_import_statement(self.offset))

    def is_called(self):
        return self.tools.word_finder.is_a_function_being_called(self.offset)

    def is_defined(self):
        return self.tools.word_finder.is_a_class_or_function_name_in_header(
            self.offset)

    def is_a_fixed_primary(self):
        return self.tools.word_finder.is_a_class_or_function_name_in_header(
            self.offset) or \
            self.tools.word_finder.is_a_name_after_from_import(self.offset)

    def is_written(self):
        return self.tools.word_finder.is_assigned_here(self.offset)

    def is_unsure(self):
        return unsure_pyname(self.get_pyname())

    def is_function_keyword_parameter(self):
        return self.tools.word_finder.is_function_keyword_parameter(
            self.offset)

    @property
    @utils.saveit
    def lineno(self):
        offset = self.get_word_range()[0]
        return self.tools.pymodule.lines.get_line_number(offset)


def same_pyname(expected, pyname):
    """Check whether `expected` and `pyname` are the same"""
    if expected is None or pyname is None:
        return False
    if expected == pyname:
        return True
    if type(expected) not in (pynames.ImportedModule, pynames.ImportedName) \
        and type(pyname) not in \
            (pynames.ImportedModule, pynames.ImportedName):
        return False
    return expected.get_definition_location() == \
        pyname.get_definition_location() and \
        expected.get_object() == pyname.get_object()


def unsure_pyname(pyname, unbound=True):
    """Return `True` if we don't know what this name references"""
    if pyname is None:
        return True
    if unbound and not isinstance(pyname, pynames.UnboundName):
        return False
    if pyname.get_object() == pyobjects.get_unknown():
        return True


class PyNameFilter(object):
    """For finding occurrences of a name."""

    def __init__(self, pyname):
        self.pyname = pyname

    def __call__(self, occurrence):
        if same_pyname(self.pyname, occurrence.get_pyname()):
            return True


class InHierarchyFilter(object):
    """Finds the occurrence if the name is in the class's hierarchy."""

    def __init__(self, pyname, implementations_only=False):
        self.pyname = pyname
        self.impl_only = implementations_only
        self.pyclass = self._get_containing_class(pyname)
        if self.pyclass is not None:
            self.name = pyname.get_object().get_name()
            self.roots = self._get_root_classes(self.pyclass, self.name)
        else:
            self.roots = None

    def __call__(self, occurrence):
        if self.roots is None:
            return
        pyclass = self._get_containing_class(occurrence.get_pyname())
        if pyclass is not None:
            roots = self._get_root_classes(pyclass, self.name)
            if self.roots.intersection(roots):
                return True

    def _get_containing_class(self, pyname):
        if isinstance(pyname, pynames.DefinedName):
            scope = pyname.get_object().get_scope()
            parent = scope.parent
            if parent is not None and parent.get_kind() == 'Class':
                return parent.pyobject

    def _get_root_classes(self, pyclass, name):
        if self.impl_only and pyclass == self.pyclass:
            return set([pyclass])
        result = set()
        for superclass in pyclass.get_superclasses():
            if name in superclass:
                result.update(self._get_root_classes(superclass, name))
        if not result:
            return set([pyclass])
        return result


class UnsureFilter(object):
    """Occurrences where we don't knoow what the name references."""

    def __init__(self, unsure):
        self.unsure = unsure

    def __call__(self, occurrence):
        if occurrence.is_unsure() and self.unsure(occurrence):
            return True


class NoImportsFilter(object):
    """Don't include import statements as occurrences."""

    def __call__(self, occurrence):
        if occurrence.is_in_import_statement():
            return False


class CallsFilter(object):
    """Filter out non-call occurrences."""

    def __call__(self, occurrence):
        if not occurrence.is_called():
            return False


class NoKeywordsFilter(object):
    """Filter out keyword parameters."""

    def __call__(self, occurrence):
        if occurrence.is_function_keyword_parameter():
            return False


class _TextualFinder(object):

    def __init__(self, name, docs=False):
        self.name = name
        self.docs = docs
        self.comment_pattern = _TextualFinder.any('comment', [r'#[^\n]*'])
        self.string_pattern = _TextualFinder.any(
            'string', [codeanalyze.get_string_pattern()])
        self.f_string_pattern = _TextualFinder.any(
            'fstring', [codeanalyze.get_formatted_string_pattern()])
        self.pattern = self._get_occurrence_pattern(self.name)

    def find_offsets(self, source):
        if not self._fast_file_query(source):
            return
        if self.docs:
            searcher = self._normal_search
        else:
            searcher = self._re_search
        for matched in searcher(source):
            yield matched

    def _re_search(self, source):
        for match in self.pattern.finditer(source):
            if match.groupdict()['occurrence']:
                yield match.start('occurrence')
            elif utils.pycompat.PY36 and match.groupdict()['fstring']:
                f_string = match.groupdict()['fstring']
                for occurrence_node in self._search_in_f_string(f_string):
                    yield match.start('fstring') + occurrence_node.col_offset

    def _search_in_f_string(self, f_string):
        tree = ast.parse(f_string)
        for node in ast.walk(tree):
            if isinstance(node, ast.Name) and node.id == self.name:
                yield node

    def _normal_search(self, source):
        current = 0
        while True:
            try:
                found = source.index(self.name, current)
                current = found + len(self.name)
                if (found == 0 or
                        not self._is_id_char(source[found - 1])) and \
                    (current == len(source) or
                        not self._is_id_char(source[current])):
                    yield found
            except ValueError:
                break

    def _is_id_char(self, c):
        return c.isalnum() or c == '_'

    def _fast_file_query(self, source):
        try:
            source.index(self.name)
            return True
        except ValueError:
            return False

    def _get_source(self, resource, pymodule):
        if resource is not None:
            return resource.read()
        else:
            return pymodule.source_code

    def _get_occurrence_pattern(self, name):
        occurrence_pattern = _TextualFinder.any('occurrence',
                                                ['\\b' + name + '\\b'])
        pattern = re.compile(occurrence_pattern + '|' + self.comment_pattern +
                             '|' + self.string_pattern + '|' +
                             self.f_string_pattern)
        return pattern

    @staticmethod
    def any(name, list_):
        return '(?P<%s>' % name + '|'.join(list_) + ')'


class _OccurrenceToolsCreator(object):

    def __init__(self, project, resource=None, pymodule=None, docs=False):
        self.project = project
        self.__resource = resource
        self.__pymodule = pymodule
        self.docs = docs

    @property
    @utils.saveit
    def name_finder(self):
        return evaluate.ScopeNameFinder(self.pymodule)

    @property
    @utils.saveit
    def source_code(self):
        return self.pymodule.source_code

    @property
    @utils.saveit
    def word_finder(self):
        return worder.Worder(self.source_code, self.docs)

    @property
    @utils.saveit
    def resource(self):
        if self.__resource is not None:
            return self.__resource
        if self.__pymodule is not None:
            return self.__pymodule.resource

    @property
    @utils.saveit
    def pymodule(self):
        if self.__pymodule is not None:
            return self.__pymodule
        return self.project.get_pymodule(self.resource)



from clikit.api.config.application_config import ApplicationConfig
from clikit.api.args.raw_args import RawArgs
from clikit.api.args.format.argument import Argument
from clikit.api.args.format.option import Option
from clikit.api.event import PRE_HANDLE
from clikit.api.event import PRE_RESOLVE
from clikit.api.event import EventDispatcher
from clikit.api.event import PreHandleEvent
from clikit.api.event import PreResolveEvent
from clikit.api.io import IO
from clikit.api.io import Input
from clikit.api.io import InputStream
from clikit.api.io import Output
from clikit.api.io import OutputStream
from clikit.api.io.flags import DEBUG
from clikit.api.io.flags import VERBOSE
from clikit.api.io.flags import VERY_VERBOSE
from clikit.api.resolver import ResolvedCommand
from clikit.formatter import AnsiFormatter
from clikit.formatter import DefaultStyleSet
from clikit.formatter import PlainFormatter
from clikit.handler.help import HelpTextHandler
from clikit.io import ConsoleIO
from clikit.io.input_stream import StandardInputStream
from clikit.io.output_stream import ErrorOutputStream
from clikit.io.output_stream import StandardOutputStream
from clikit.resolver.default_resolver import DefaultResolver
from clikit.resolver.help_resolver import HelpResolver
from clikit.ui.components import NameVersion


class DefaultApplicationConfig(ApplicationConfig):
    """
    The default application configuration.
    """

    def configure(self):
        self.set_io_factory(self.create_io)
        self.add_event_listener(PRE_RESOLVE, self.resolve_help_command)
        self.add_event_listener(PRE_HANDLE, self.print_version)

        self.add_option("help", "h", Option.NO_VALUE, "Display this help message")
        self.add_option("quiet", "q", Option.NO_VALUE, "Do not output any message")
        self.add_option(
            "verbose",
            "v",
            Option.OPTIONAL_VALUE,
            "Increase the verbosity of messages: "
            '"-v" for normal output, '
            '"-vv" for more verbose output '
            'and "-vvv" for debug',
        )
        self.add_option(
            "version", "V", Option.NO_VALUE, "Display this application version"
        )
        self.add_option("ansi", None, Option.NO_VALUE, "Force ANSI output")
        self.add_option("no-ansi", None, Option.NO_VALUE, "Disable ANSI output")
        self.add_option(
            "no-interaction",
            "n",
            Option.NO_VALUE,
            "Do not ask any interactive question",
        )

        with self.command("help") as c:
            c.default()
            c.set_description("Display the manual of a command")
            c.add_argument(
                "command", Argument.OPTIONAL | Argument.MULTI_VALUED, "The command name"
            )
            c.set_handler(HelpTextHandler(HelpResolver()))

    def create_io(
        self,
        application,
        args,
        input_stream=None,
        output_stream=None,
        error_stream=None,
    ):  # type: (Application, RawArgs, InputStream, OutputStream, OutputStream) -> IO
        if input_stream is None:
            input_stream = StandardInputStream()

        if output_stream is None:
            output_stream = StandardOutputStream()

        if error_stream is None:
            error_stream = ErrorOutputStream()

        style_set = application.config.style_set

        if args.has_option_token("--no-ansi"):
            output_formatter = error_formatter = PlainFormatter(style_set)
        elif args.has_option_token("--ansi"):
            output_formatter = error_formatter = AnsiFormatter(style_set, True)
        else:
            if output_stream.supports_ansi():
                output_formatter = AnsiFormatter(style_set)
            else:
                output_formatter = PlainFormatter(style_set)

            if error_stream.supports_ansi():
                error_formatter = AnsiFormatter(style_set)
            else:
                error_formatter = PlainFormatter(style_set)

        io = self.io_class(
            Input(input_stream),
            Output(output_stream, output_formatter),
            Output(error_stream, error_formatter),
        )

        if args.has_option_token("-vvv") or self.is_debug():
            io.set_verbosity(DEBUG)
        elif args.has_option_token("-vv"):
            io.set_verbosity(VERY_VERBOSE)
        elif args.has_option_token("-v"):
            io.set_verbosity(VERBOSE)

        if args.has_option_token("--quiet") or args.has_option_token("-q"):
            io.set_quiet(True)

        if args.has_option_token("--no-interaction") or args.has_option_token("-n"):
            io.set_interactive(False)

        return io

    @property
    def io_class(self):  # type: () -> IO.__class__
        return ConsoleIO

    @property
    def default_style_set(self):  # type: () -> DefaultStyleSet
        return DefaultStyleSet()

    @property
    def default_command_resolver(self):  # type: () -> DefaultResolver
        return DefaultResolver()

    def resolve_help_command(
        self, event, event_name, dispatcher
    ):  # type: (PreResolveEvent, str, EventDispatcher) -> None
        args = event.raw_args
        application = event.application

        if args.has_option_token("-h") or args.has_option_token("--help"):
            command = application.get_command("help")

            # Enable lenient parsing
            parsed_args = command.parse(args, True)

            event.set_resolved_command(ResolvedCommand(command, parsed_args))
            event.stop_propagation()

    def print_version(
        self, event, event_name, dispatcher
    ):  # type: (PreHandleEvent, str, EventDispatcher) -> None
        if event.args.is_option_set("version"):
            version = NameVersion(event.command.application.config)
            version.render(event.io)

            event.handled(True)



# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2014 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from datetime import datetime
import xml.etree.ElementTree

import requests

from .exceptions import HTTPError


def _get(pypi_server):
    """
    Query the PyPI RSS feed and return a list
    of XML items.
    """
    response = requests.get(pypi_server)
    if response.status_code >= 300:
        raise HTTPError(status_code=response.status_code,
                        reason=response.reason)
    if hasattr(response.content, 'decode'):
        tree = xml.etree.ElementTree.fromstring(response.content.decode())
    else:
        tree = xml.etree.ElementTree.fromstring(response.content)
    channel = tree.find('channel')
    return channel.findall('item')


def newest_packages(
        pypi_server="https://pypi.python.org/pypi?%3Aaction=packages_rss"):
    """
    Constructs a request to the PyPI server and returns a list of
    :class:`yarg.parse.Package`.

    :param pypi_server: (option) URL to the PyPI server.

        >>> import yarg
        >>> yarg.newest_packages()
        [<Package yarg>, <Package gray>, <Package ragy>]
    """
    items = _get(pypi_server)
    i = []
    for item in items:
        i_dict = {'name': item[0].text.split()[0],
                  'url': item[1].text,
                  'description': item[3].text,
                  'date': item[4].text}
        i.append(Package(i_dict))
    return i


def latest_updated_packages(
        pypi_server="https://pypi.python.org/pypi?%3Aaction=rss"):
    """
    Constructs a request to the PyPI server and returns a list of
    :class:`yarg.parse.Package`.

    :param pypi_server: (option) URL to the PyPI server.

        >>> import yarg
        >>> yarg.latest_updated_packages()
        [<Package yarg>, <Package gray>, <Package ragy>]
    """
    items = _get(pypi_server)
    i = []
    for item in items:
        name, version = item[0].text.split()
        i_dict = {'name': name,
                  'version': version,
                  'url': item[1].text,
                  'description': item[2].text,
                  'date': item[3].text}
        i.append(Package(i_dict))
    return i


class Package(object):
    """
    A PyPI package generated from the RSS feed information.

    :param pypi_dict: A dictionary retrieved from the PyPI server.
    """

    def __init__(self, pypi_dict):
        self._content = pypi_dict

    def __repr__(self):
        return "<Package {0}>".format(self.name)

    @property
    def name(self):
        """
            >>> package = yarg.newest_packages()[0]
            >>> package.name
            u'yarg'
            >>> package = yarg.latest_updated_packages()[0]
            >>> package.name
            u'yarg'
        """
        return self._content['name']

    @property
    def version(self):
        """
            >>> package = yarg.newest_packages()[0]
            >>> package.name
            u'yarg'
            >>> package = yarg.latest_updated_packages()[0]
            >>> package.name
            u'yarg'
        """
        if 'version' not in self._content:
            return None
        return self._content['version']

    @property
    def url(self):
        """
        This is only available for :meth:`yarg.latest_updated_packages`, for
        :meth:`yarg.newest_packages` will return `None`

            >>> package = yarg.latest_updated_packages()[0]
            >>> package.url
            u'http://pypi.python.org/pypi/yarg'
        """
        return self._content['url']

    @property
    def date(self):
        """
            >>> package = yarg.newest_packages()[0]
            >>> package.date
            datetime.datetime(2014, 8, 9, 8, 40, 20)
            >>> package = yarg.latest_updated_packages()[0]
            >>> package.date
            datetime.datetime(2014, 8, 9, 8, 40, 20)
        """
        return datetime.strptime(self._content['date'],
                                 "%d %b %Y %H:%M:%S %Z")

    @property
    def description(self):
        """
            >>> package = yarg.newest_packages()[0]
            >>> package.description
            u'Some random summary stuff'
            >>> package = yarg.latest_updated_packages()[0]
            >>> package.description
            u'Some random summary stuff'
        """
        return self._content['description']



import sys


PY2 = sys.version_info[0] == 2

if PY2:
    range = xrange


def wf_levenshtein(string_1, string_2):
    """
    Calculates the Levenshtein distance between two strings.

    This version uses the Wagner-Fischer algorithm.

    Usage::

        >>> wf_levenshtein('kitten', 'sitting')
        3
        >>> wf_levenshtein('kitten', 'kitten')
        0
        >>> wf_levenshtein('', '')
        0

    """
    len_1 = len(string_1) + 1
    len_2 = len(string_2) + 1

    d = [0] * (len_1 * len_2)

    for i in range(len_1):
        d[i] = i
    for j in range(len_2):
        d[j * len_1] = j

    for j in range(1, len_2):
        for i in range(1, len_1):
            if string_1[i - 1] == string_2[j - 1]:
                d[i + j * len_1] = d[i - 1 + (j - 1) * len_1]
            else:
                d[i + j * len_1] = min(
                    d[i - 1 + j * len_1] + 1,  # deletion
                    d[i + (j - 1) * len_1] + 1,  # insertion
                    d[i - 1 + (j - 1) * len_1] + 1,  # substitution
                )

    return d[-1]


def wfi_levenshtein(string_1, string_2):
    """
    Calculates the Levenshtein distance between two strings.

    This version uses an iterative version of the Wagner-Fischer algorithm.

    Usage::

        >>> wfi_levenshtein('kitten', 'sitting')
        3
        >>> wfi_levenshtein('kitten', 'kitten')
        0
        >>> wfi_levenshtein('', '')
        0

    """
    if string_1 == string_2:
        return 0

    len_1 = len(string_1)
    len_2 = len(string_2)

    if len_1 == 0:
        return len_2
    if len_2 == 0:
        return len_1

    if len_1 > len_2:
        string_2, string_1 = string_1, string_2
        len_2, len_1 = len_1, len_2

    d0 = [i for i in range(len_2 + 1)]
    d1 = [j for j in range(len_2 + 1)]

    for i in range(len_1):
        d1[0] = i + 1
        for j in range(len_2):
            cost = d0[j]

            if string_1[i] != string_2[j]:
                # substitution
                cost += 1

                # insertion
                x_cost = d1[j] + 1
                if x_cost < cost:
                    cost = x_cost

                # deletion
                y_cost = d0[j + 1] + 1
                if y_cost < cost:
                    cost = y_cost

            d1[j + 1] = cost

        d0, d1 = d1, d0

    return d0[-1]



'''
A simpler setup version just to compile the speedup module.

It should be used as:

python setup_cython build_ext --inplace

Note: the .c file and other generated files are regenerated from
the .pyx file by running "python build_tools/build.py"
'''

import os
import sys
from setuptools import setup

os.chdir(os.path.dirname(os.path.abspath(__file__)))

IS_PY36_OR_GREATER = sys.version_info > (3, 6)


def process_args():
    extension_folder = None
    target_pydevd_name = None
    target_frame_eval = None
    force_cython = False

    for i, arg in enumerate(sys.argv[:]):
        if arg == '--build-lib':
            extension_folder = sys.argv[i + 1]
            # It shouldn't be removed from sys.argv (among with --build-temp) because they're passed further to setup()
        if arg.startswith('--target-pyd-name='):
            sys.argv.remove(arg)
            target_pydevd_name = arg[len('--target-pyd-name='):]
        if arg.startswith('--target-pyd-frame-eval='):
            sys.argv.remove(arg)
            target_frame_eval = arg[len('--target-pyd-frame-eval='):]
        if arg == '--force-cython':
            sys.argv.remove(arg)
            force_cython = True

    return extension_folder, target_pydevd_name, target_frame_eval, force_cython


def process_template_lines(template_lines):
    # Create 2 versions of the template, one for Python 3.8 and another for Python 3.9
    for version in ('38', '39'):
        yield '### WARNING: GENERATED CODE, DO NOT EDIT!'
        yield '### WARNING: GENERATED CODE, DO NOT EDIT!'
        yield '### WARNING: GENERATED CODE, DO NOT EDIT!'

        for line in template_lines:
            if version == '38':
                line = line.replace('get_bytecode_while_frame_eval(PyFrameObject * frame_obj, int exc)', 'get_bytecode_while_frame_eval_38(PyFrameObject * frame_obj, int exc)')
                line = line.replace('CALL_EvalFrameDefault', 'CALL_EvalFrameDefault_38(frame_obj, exc)')
            else:  # 3.9
                line = line.replace('get_bytecode_while_frame_eval(PyFrameObject * frame_obj, int exc)', 'get_bytecode_while_frame_eval_39(PyThreadState* tstate, PyFrameObject * frame_obj, int exc)')
                line = line.replace('CALL_EvalFrameDefault', 'CALL_EvalFrameDefault_39(tstate, frame_obj, exc)')

            yield line

        yield '### WARNING: GENERATED CODE, DO NOT EDIT!'
        yield '### WARNING: GENERATED CODE, DO NOT EDIT!'
        yield '### WARNING: GENERATED CODE, DO NOT EDIT!'
        yield ''
        yield ''


def process_template_file(contents):
    ret = []
    template_lines = []

    append_to = ret
    for line in contents.splitlines(keepends=False):
        if line.strip() == '### TEMPLATE_START':
            append_to = template_lines
        elif line.strip() == '### TEMPLATE_END':
            append_to = ret
            for line in process_template_lines(template_lines):
                ret.append(line)
        else:
            append_to.append(line)

    return '\n'.join(ret)


def build_extension(dir_name, extension_name, target_pydevd_name, force_cython, extended=False, has_pxd=False, template=False):
    pyx_file = os.path.join(os.path.dirname(__file__), dir_name, "%s.pyx" % (extension_name,))

    if template:
        pyx_template_file = os.path.join(os.path.dirname(__file__), dir_name, "%s.template.pyx" % (extension_name,))
        with open(pyx_template_file, 'r') as stream:
            contents = stream.read()

        contents = process_template_file(contents)

        with open(pyx_file, 'w') as stream:
            stream.write(contents)

    if target_pydevd_name != extension_name:
        # It MUST be there in this case!
        # (otherwise we'll have unresolved externals because the .c file had another name initially).
        import shutil

        # We must force cython in this case (but only in this case -- for the regular setup in the user machine, we
        # should always compile the .c file).
        force_cython = True

        new_pyx_file = os.path.join(os.path.dirname(__file__), dir_name, "%s.pyx" % (target_pydevd_name,))
        new_c_file = os.path.join(os.path.dirname(__file__), dir_name, "%s.c" % (target_pydevd_name,))
        shutil.copy(pyx_file, new_pyx_file)
        pyx_file = new_pyx_file
        if has_pxd:
            pxd_file = os.path.join(os.path.dirname(__file__), dir_name, "%s.pxd" % (extension_name,))
            new_pxd_file = os.path.join(os.path.dirname(__file__), dir_name, "%s.pxd" % (target_pydevd_name,))
            shutil.copy(pxd_file, new_pxd_file)
        assert os.path.exists(pyx_file)

    try:
        c_files = [os.path.join(dir_name, "%s.c" % target_pydevd_name), ]
        if force_cython:
            for c_file in c_files:
                try:
                    os.remove(c_file)
                except:
                    pass
            from Cython.Build import cythonize  # @UnusedImport
            # Generate the .c files in cythonize (will not compile at this point).

            target = "%s/%s.pyx" % (dir_name, target_pydevd_name,)
            cythonize([target])

            # Workarounds needed in CPython 3.8 and 3.9 to access PyInterpreterState.eval_frame.
            for c_file in c_files:
                with open(c_file, 'r') as stream:
                    c_file_contents = stream.read()

                if '#include "internal/pycore_gc.h"' not in c_file_contents:
                    c_file_contents = c_file_contents.replace('#include "Python.h"', '''#include "Python.h"
#if PY_VERSION_HEX >= 0x03090000
#include "internal/pycore_gc.h"
#include "internal/pycore_interp.h"
#endif
''')

                if '#include "internal/pycore_pystate.h"' not in c_file_contents:
                    c_file_contents = c_file_contents.replace('#include "pystate.h"', '''#include "pystate.h"
#if PY_VERSION_HEX >= 0x03080000
#include "internal/pycore_pystate.h"
#endif
''')

                # We want the same output on Windows and Linux.
                c_file_contents = c_file_contents.replace('\r\n', '\n').replace('\r', '\n')
                c_file_contents = c_file_contents.replace(r'_pydevd_frame_eval\\release_mem.h', '_pydevd_frame_eval/release_mem.h')
                c_file_contents = c_file_contents.replace(r'_pydevd_frame_eval\\pydevd_frame_evaluator.pyx', '_pydevd_frame_eval/pydevd_frame_evaluator.pyx')
                c_file_contents = c_file_contents.replace(r'_pydevd_bundle\\pydevd_cython.pxd', '_pydevd_bundle/pydevd_cython.pxd')
                c_file_contents = c_file_contents.replace(r'_pydevd_bundle\\pydevd_cython.pyx', '_pydevd_bundle/pydevd_cython.pyx')

                with open(c_file, 'w') as stream:
                    stream.write(c_file_contents)

        # Always compile the .c (and not the .pyx) file (which we should keep up-to-date by running build_tools/build.py).
        from distutils.extension import Extension
        ext_modules = [Extension("%s%s.%s" % (dir_name, "_ext" if extended else "", target_pydevd_name,),
                                 c_files,
                                 # uncomment to generate pdbs for visual studio.
                                 # extra_compile_args=["-Zi", "/Od"],
                                 # extra_link_args=["-debug"],
                                 )]

        # This is needed in CPython 3.8 to be able to include internal/pycore_pystate.h
        # (needed to set PyInterpreterState.eval_frame).
        for module in ext_modules:
            module.define_macros = [('Py_BUILD_CORE_MODULE', '1')]
        setup(
            name='Cythonize',
            ext_modules=ext_modules
        )
    finally:
        if target_pydevd_name != extension_name:
            try:
                os.remove(new_pyx_file)
            except:
                import traceback
                traceback.print_exc()
            try:
                os.remove(new_c_file)
            except:
                import traceback
                traceback.print_exc()
            if has_pxd:
                try:
                    os.remove(new_pxd_file)
                except:
                    import traceback
                    traceback.print_exc()


extension_folder, target_pydevd_name, target_frame_eval, force_cython = process_args()

extension_name = "pydevd_cython"
if target_pydevd_name is None:
    target_pydevd_name = extension_name
build_extension("_pydevd_bundle", extension_name, target_pydevd_name, force_cython, extension_folder, True)

if IS_PY36_OR_GREATER:
    extension_name = "pydevd_frame_evaluator"
    if target_frame_eval is None:
        target_frame_eval = extension_name
    build_extension("_pydevd_frame_eval", extension_name, target_frame_eval, force_cython, extension_folder, True, template=True)

if extension_folder:
    os.chdir(extension_folder)
    for folder in [file for file in os.listdir(extension_folder) if
                   file != 'build' and os.path.isdir(os.path.join(extension_folder, file))]:
        file = os.path.join(folder, "__init__.py")
        if not os.path.exists(file):
            open(file, 'a').close()



from typing import List
from typing import Optional


class CommandName(object):
    """
    A command name in the console arguments.
    """

    def __init__(
        self, string, aliases=None
    ):  # type: (str, Optional[List[str]]) -> None
        if aliases is None:
            aliases = []

        self._string = string
        self._aliases = aliases

    @property
    def string(self):  # type: () -> str
        return self._string

    @property
    def aliases(self):  # type: () -> List[str]
        return self._aliases

    def match(self, string):  # type: (str) -> bool
        return self._string == string or string in self._aliases

    def __str__(self):  # type: () -> str
        return self._string

    def __repr__(self):  # type: () -> str
        return 'CommandName("{}")'.format(self._string)



import os
import traceback
from _pydevd_bundle.pydevd_constants import ForkSafeLock

_pid = os.getpid()
_pid_msg = '%s: ' % (_pid,)

_debug_lock = ForkSafeLock()

DEBUG = False
DEBUG_FILE = os.path.join(os.path.dirname(__file__), '__debug_output__.txt')


def debug(msg):
    if DEBUG:
        with _debug_lock:
            _pid_prefix = _pid_msg
            if isinstance(msg, bytes):
                _pid_prefix = _pid_prefix.encode('utf-8')

                if not msg.endswith(b'\r') and not msg.endswith(b'\n'):
                    msg += b'\n'
                mode = 'a+b'
            else:
                if not msg.endswith('\r') and not msg.endswith('\n'):
                    msg += '\n'
                mode = 'a+'
            with open(DEBUG_FILE, mode) as stream:
                stream.write(_pid_prefix)
                stream.write(msg)


def debug_exception(msg=None):
    if DEBUG:
        if msg:
            debug(msg)

        with _debug_lock:

            with open(DEBUG_FILE, 'a+') as stream:
                _pid_prefix = _pid_msg
                if isinstance(msg, bytes):
                    _pid_prefix = _pid_prefix.encode('utf-8')
                stream.write(_pid_prefix)

                traceback.print_exc(file=stream)



# -*- coding: utf-8 -*-


class Serializer(object):
    """
    Abstract serializer.
    """

    def serialize(self, data):
        """
        Serialize data.

        :param data: The data to serialize
        :type data: mixed

        :rtype: str
        """
        raise NotImplementedError()

    def unserialize(self, data):
        """
        Unserialize data.

        :param data: The data to unserialize
        :type data: mixed

        :rtype: str
        """
        raise NotImplementedError()



from __future__ import absolute_import, unicode_literals

import os

from virtualenv.util.path import Path
from virtualenv.util.six import ensure_text

from ..via_template import ViaTemplateActivator


class NushellActivator(ViaTemplateActivator):
    def templates(self):
        yield Path("activate.nu")
        yield Path("deactivate.nu")

    def replacements(self, creator, dest_folder):
        # Due to nushell scoping, it isn't easy to create a function that will
        # deactivate the environment. For that reason a __DEACTIVATE_PATH__
        # replacement pointing to the deactivate.nu file is created

        return {
            "__VIRTUAL_PROMPT__": "" if self.flag_prompt is None else self.flag_prompt,
            "__VIRTUAL_ENV__": ensure_text(str(creator.dest)),
            "__VIRTUAL_NAME__": creator.env_name,
            "__BIN_NAME__": ensure_text(str(creator.bin_dir.relative_to(creator.dest))),
            "__PATH_SEP__": ensure_text(os.pathsep),
            "__DEACTIVATE_PATH__": ensure_text(str(Path(dest_folder) / "deactivate.nu")),
        }



# -*- coding: utf-8 -*-

try:
    import msgpack
except ImportError:
    msgpack = None

from .serializer import Serializer


class MsgPackSerializer(Serializer):
    """
    Serializer that uses `msgpack <https://pypi.python.org/pypi/msgpack-python/>`_ representations.

    By default, this serializer does not support serializing custom objects.
    """

    def serialize(self, data):
        """
        Serialize data.

        :param data: The data to serialize
        :type data: mixed

        :rtype: str
        """
        return msgpack.packb(data, use_bin_type=True)

    def unserialize(self, data):
        """
        Unserialize data.

        :param data: The data to unserialize
        :type data: mixed

        :rtype: str
        """
        return msgpack.unpackb(data, encoding='utf-8')



# -*- coding: utf-8 -*-

import math
import datetime
import types
import hashlib
from functools import wraps
from .contracts.repository import Repository as CacheContract
from .helpers import value
from .utils import encode, decode


class Repository(CacheContract):

    _default = 60

    def __init__(self, store):
        """
        :param store: The underlying cache store
        :type store: Store
        """
        self._store = store

    def has(self, key):
        """
        Determine if an item exists in the cache.

        :param key: The cache key
        :type key: str

        :rtype: bool
        """
        return self.get(key) is not None

    def get(self, key, default=None):
        """
        Retrieve an item from the cache by key.

        :param key: The cache key
        :type key: str

        :param default: The default value to return
        :type default: mixed

        :rtype: mixed
        """
        val = self._store.get(key)

        if val is None:
            return value(default)

        return val

    def pull(self, key, default=None):
        """
        Retrieve an item from the cache by key and delete ir.

        :param key: The cache key
        :type key: str

        :param default: The default value to return
        :type default: mixed

        :rtype: mixed
        """
        val = self.get(key, default)

        self.forget(key)

        return val

    def put(self, key, val, minutes):
        """
        Store an item in the cache.

        :param key: The cache key
        :type key: str

        :param val: The cache value
        :type val: mixed

        :param minutes: The lifetime in minutes of the cached value
        :type minutes: int|datetime
        """
        minutes = self._get_minutes(minutes)

        if minutes is not None:
            self._store.put(key, val, minutes)

    def add(self, key, val, minutes):
        """
        Store an item in the cache if it does not exist.

        :param key: The cache key
        :type key: str

        :param val: The cache value
        :type val: mixed

        :param minutes: The lifetime in minutes of the cached value
        :type minutes: int|datetime

        :rtype: bool
        """
        if hasattr(self._store, 'add'):
            return self._store.add(key, val, self._get_minutes(minutes))

        if not self.has(key):
            self.put(key, val, minutes)

            return True

        return False

    def forever(self, key, val):
        """
        Store an item in the cache indefinitely.

        :param key: The cache key
        :type key: str

        :param val: The cache value
        :type val: mixed
        """
        self._store.forever(key, val)

    def remember(self, key, minutes, callback):
        """
        Get an item from the cache, or store the default value.

        :param key: The cache key
        :type key: str

        :param minutes: The lifetime in minutes of the cached value
        :type minutes: int or datetime

        :param callback: The default function
        :type callback: mixed

        :rtype: mixed
        """
        # If the item exists in the cache we will just return this immediately
        # otherwise we will execute the given callback and cache the result
        # of that execution for the given number of minutes in storage.
        val = self.get(key)
        if val is not None:
            return val

        val = value(callback)

        self.put(key, val, minutes)

        return val

    def remember_forever(self, key, callback):
        """
        Get an item from the cache, or store the default value forever.

        :param key: The cache key
        :type key: str

        :param callback: The default function
        :type callback: mixed

        :rtype: mixed
        """
        # If the item exists in the cache we will just return this immediately
        # otherwise we will execute the given callback and cache the result
        # of that execution forever.
        val = self.get(key)
        if val is not None:
            return val

        val = value(callback)

        self.forever(key, val)

        return val

    def forget(self, key):
        """
        Remove an item from the cache.

        :param key: The cache key
        :type key: str

        :rtype: bool
        """
        success = self._store.forget(key)

        return success

    def get_default_cache_time(self):
        """
        Get the default cache time.

        :rtype: int
        """
        return self._default

    def set_default_cache_time(self, minutes):
        """
        Set the default cache time.

        :param minutes: The default cache time
        :type minutes: int

        :rtype: self
        """
        self._default = minutes

        return self

    def get_store(self):
        """
        Get the cache store implementation.

        :rtype: Store
        """
        return self._store

    def __getitem__(self, item):
        return self.get(item)

    def __setitem__(self, key, val):
        self.put(key, val, self._default)

    def __delitem__(self, key):
        self.forget(key)

    def _get_minutes(self, duration):
        """
        Calculate the number of minutes with the given duration.

        :param duration: The duration
        :type duration: int or datetime

        :rtype: int or None
        """
        if isinstance(duration, datetime.datetime):
            from_now = (duration - datetime.datetime.now()).total_seconds()
            from_now = math.ceil(from_now / 60)

            if from_now > 0:
                return from_now

            return

        return duration

    def _hash(self, value):
        """
        Calculate the hash given a value.

        :param value: The value to hash
        :type value: str or bytes

        :rtype: str
        """
        return hashlib.sha1(encode(value)).hexdigest()

    def _get_key(self, fn, args, kwargs):
        """
        Calculate a cache key given a function, args and kwargs.

        :param fn: The function
        :type fn: callable or str

        :param args: The function args
        :type args: tuple

        :param kwargs: The function kwargs
        :type kwargs: dict

        :rtype: str
        """
        if args:
            serialized_arguments = (
                self._store.serialize(args[1:])
                + self._store.serialize([(k, kwargs[k]) for k in sorted(kwargs.keys())])
            )
        else:
            serialized_arguments = self._store.serialize([(k, kwargs[k]) for k in sorted(kwargs.keys())])

        if isinstance(fn, types.MethodType):
            key = self._hash('%s.%s.%s'
                             % (fn.__self__.__class__.__name__,
                                args[0].__name__,
                                serialized_arguments))
        elif isinstance(fn, types.FunctionType):
            key = self._hash('%s.%s'
                             % (fn.__name__,
                                serialized_arguments))
        else:
            key = '%s:' % fn + self._hash(serialized_arguments)

        return key

    def __getattr__(self, item):
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            return getattr(self._store, item)

    def __call__(self, *args, **kwargs):
        if args and isinstance(args[0], (types.FunctionType, types.MethodType)):
            fn = args[0]

            @wraps(fn)
            def wrapper(*a, **kw):
                return self.remember(
                    self._get_key(fn, a, kw),
                    self._default,
                    lambda: fn(*a, **kw)
                )

            return wrapper
        else:
            k = kwargs.get('key')
            minutes = kwargs.get('minutes', self._default)

            def decorated(fn):
                key = k

                @wraps(fn)
                def wrapper(*a, **kw):
                    return self.remember(
                        self._get_key(key or fn, a, kw),
                        minutes,
                        lambda: fn(*a, **kw)
                    )

                return wrapper

            return decorated



from clikit.api.io import IO
from clikit.ui import Component


class EmptyLine(Component):
    """
    An empty line.
    """

    def render(self, io, indentation=0):  # type: (IO, int) -> None
        io.write("\n")



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from ..exceptions import NoSuchArgumentException
from ..exceptions import NoSuchOptionException

from .argument import Argument
from .option import Option
from .command_name import CommandName
from .command_option import CommandOption


class ArgsFormat(object):
    """
    The format used to parse a RawArgs instance.
    """

    def __init__(
        self, elements=None, base_format=None
    ):  # type: (Optional[Union[List[Any], ArgsFormatBuilder]], Optional[ArgsFormat])
        from .args_format_builder import ArgsFormatBuilder

        if elements is None:
            elements = []

        if isinstance(elements, ArgsFormatBuilder):
            builder = elements
        else:
            builder = self._create_builder_for_elements(elements)

        if base_format is None:
            base_format = builder.base_format

        self._base_format = base_format
        self._command_names = builder.get_command_names(False)
        self._command_options = {}
        self._command_options_by_short_name = {}
        self._arguments = builder.get_arguments(False)
        self._options = builder.get_options(False)
        self._options_by_short_name = {}
        self._has_multi_valued_arg = builder.has_multi_valued_argument(False)
        self._hash_optional_arg = builder.has_optional_argument(False)

        for option in self._options.values():
            if option.short_name:
                self._options_by_short_name[option.short_name] = option

        for command_option in builder.get_command_options():
            self._command_options[command_option.long_name] = command_option

            if command_option.short_name:
                self._command_options_by_short_name[
                    command_option.short_name
                ] = command_option

            for long_alias in command_option.long_aliases:
                self._command_options[long_alias] = command_option

            for short_alias in command_option.short_aliases:
                self._command_options_by_short_name[short_alias] = command_option

    @property
    def base_format(self):  # type: () -> ArgsFormat
        return self._base_format

    def has_command_names(self, include_base=True):  # type: (bool) -> bool
        if self._command_names:
            return True

        if include_base and self._base_format:
            return self._base_format.has_command_names()

        return False

    def get_command_names(self, include_base=True):  # type: (bool) -> List[CommandName]
        command_names = self._command_names

        if include_base and self._base_format:
            command_names = self._base_format.get_command_names() + command_names

        return command_names

    def has_command_option(self, name, include_base=True):  # type: (str, bool) -> bool
        if name in self._command_options or name in self._command_options_by_short_name:
            return True

        if include_base and self._base_format:
            return self._base_format.has_command_option(name)

        return False

    def has_command_options(self, include_base=True):  # type: (bool) -> bool
        if self._command_options:
            return True

        if include_base and self._base_format:
            return self._base_format.has_command_options()

        return False

    def get_command_option(
        self, name, include_base=True
    ):  # type: (str, bool) -> CommandOption
        if name in self._command_options:
            return self._command_options[name]

        if name in self._command_options_by_short_name:
            return self._command_options_by_short_name[name]

        if include_base and self._base_format:
            return self._base_format.get_command_option(name)

        raise NoSuchOptionException(name)

    def get_command_options(
        self, include_base=True
    ):  # type: (bool) -> List[CommandOption]
        command_options = list(self._command_options.values())

        if include_base and self._base_format:
            command_options += self._base_format.get_command_options()

        return command_options

    def has_argument(
        self, name, include_base=True
    ):  # type: (Union[str, int], bool) -> bool
        arguments = self.get_arguments(include_base)

        if isinstance(name, int):
            return name < len(arguments)

        return name in arguments

    def has_multi_valued_argument(self, include_base=True):  # type: (bool) -> bool
        if self._has_multi_valued_arg:
            return True

        if include_base and self._base_format:
            return self._base_format.has_multi_valued_argument()

        return False

    def has_optional_argument(self, include_base=True):  # type: (bool) -> bool
        if self._hash_optional_arg:
            return True

        if include_base and self._base_format:
            return self._base_format.has_optional_argument()

        return False

    def has_required_argument(self, include_base=True):  # type: (bool) -> bool
        if not self._hash_optional_arg and self._arguments:
            return True

        if include_base and self._base_format:
            return self._base_format.has_required_argument()

        return False

    def has_arguments(self, include_base=True):  # type: (bool) -> bool
        if self._arguments:
            return True

        if include_base and self._base_format:
            return self._base_format.has_arguments()

        return False

    def get_argument(
        self, name, include_base=True
    ):  # type: (Union[str, int], bool) -> Argument
        if isinstance(name, int):
            arguments = list(self.get_arguments(include_base).values())

            if name >= len(arguments):
                raise NoSuchArgumentException(name)
        else:
            arguments = self.get_arguments(include_base)

            if name not in arguments:
                raise NoSuchArgumentException(name)

        return arguments[name]

    def get_arguments(self, include_base=True):  # type: (bool) -> Dict[str, Argument]
        arguments = self._arguments.copy()

        if include_base and self._base_format:
            base_arguments = self._base_format.get_arguments()
            base_arguments.update(arguments)
            arguments = base_arguments

        return arguments

    def has_option(self, name, include_base=True):  # type: (str, bool) -> bool
        if name in self._options or name in self._options_by_short_name:
            return True

        if include_base and self._base_format:
            return self._base_format.has_option(name)

        return False

    def has_options(self, include_base=True):  # type: (bool) -> bool
        if self._options:
            return True

        if include_base and self._base_format:
            return self._base_format.has_options()

        return False

    def get_option(self, name, include_base=True):  # type: (str, bool) -> Option
        if name in self._options:
            return self._options[name]

        if name in self._options_by_short_name:
            return self._options_by_short_name[name]

        if include_base and self._base_format:
            return self._base_format.get_option(name)

        raise NoSuchOptionException(name)

    def get_options(self, include_base=True):  # type: (bool) -> Dict[str, Option]
        options = self._options.copy()

        if include_base and self._base_format:
            options.update(self._base_format.get_options())

        return options

    def _create_builder_for_elements(
        self, elements, base_format=None
    ):  # type: (List[Any], Optional[ArgsFormat]) -> ArgsFormatBuilder
        from .args_format_builder import ArgsFormatBuilder

        builder = ArgsFormatBuilder(base_format)

        for element in elements:
            if isinstance(element, CommandName):
                builder.add_command_name(element)
            elif isinstance(element, CommandOption):
                builder.add_command_option(element)
            elif isinstance(element, Option):
                builder.add_option(element)
            elif isinstance(element, Argument):
                builder.add_argument(element)

        return builder



# -*- coding: utf-8 -*-

import threading
import types
from .contracts.factory import Factory
from .contracts.store import Store

from .stores import (
    DictStore,
    FileStore,
    RedisStore,
    MemcachedStore
)

from .repository import Repository
from .serializers import (
    Serializer,
    JsonSerializer,
    MsgPackSerializer,
    PickleSerializer
)


class CacheManager(Factory, threading.local):
    """
    A CacheManager is a pool of cache stores.
    """

    _serializers = {
        'json': JsonSerializer(),
        'msgpack': MsgPackSerializer(),
        'pickle': PickleSerializer()
    }

    def __init__(self, config):
        super(CacheManager, self).__init__()

        self._config = config
        self._stores = {}
        self._custom_creators = {}
        self._serializer = self._resolve_serializer(config.get('serializer', 'pickle'))

    def store(self, name=None):
        """
        Get a cache store instance by name.

        :param name: The cache store name
        :type name: str

        :rtype: Repository
        """
        if name is None:
            name = self.get_default_driver()

        self._stores[name] = self._get(name)

        return self._stores[name]

    def driver(self, name=None):
        """
        Get a cache store instance by name.

        :param name: The cache store name
        :type name: str

        :rtype: Repository
        """
        return self.store(name)

    def _get(self, name):
        """
        Attempt to get the store from the local cache.

        :param name: The store name
        :type name: str

        :rtype: Repository
        """
        return self._stores.get(name, self._resolve(name))

    def _resolve(self, name):
        """
        Resolve the given store

        :param name: The store to resolve
        :type name: str

        :rtype: Repository
        """
        config = self._get_config(name)

        if not config:
            raise RuntimeError('Cache store [%s] is not defined.' % name)

        if config['driver'] in self._custom_creators:
            repository = self._call_custom_creator(config)
        else:
            repository = getattr(self, '_create_%s_driver' % config['driver'])(config)

        if 'serializer' in config:
            serializer = self._resolve_serializer(config['serializer'])
        else:
            serializer = self._serializer

        repository.get_store().set_serializer(serializer)

        return repository

    def _call_custom_creator(self, config):
        """
        Call a custom driver creator.

        :param config: The driver configuration
        :type config: dict

        :rtype: Repository
        """
        creator = self._custom_creators[config['driver']](config)

        if isinstance(creator, Store):
            creator = self.repository(creator)

        if not isinstance(creator, Repository):
            raise RuntimeError('Custom creator should return a Repository instance.')

        return creator

    def _create_dict_driver(self, config):
        """
        Create an instance of the dict cache driver.

        :param config: The driver configuration
        :type config: dict

        :rtype: Repository
        """
        return self.repository(DictStore())

    def _create_file_driver(self, config):
        """
        Create an instance of the file cache driver.

        :param config: The driver configuration
        :type config: dict

        :rtype: Repository
        """
        kwargs = {
            'directory': config['path']
        }

        if 'hash_type' in config:
            kwargs['hash_type'] = config['hash_type']

        return self.repository(FileStore(**kwargs))

    def _create_redis_driver(self, config):
        """
        Create an instance of the redis cache driver.

        :param config: The driver configuration
        :type config: dict

        :return: Repository
        """
        return self.repository(RedisStore(**config))

    def _create_memcached_driver(self, config):
        """
        Create an instance of the redis cache driver.

        :param config: The driver configuration
        :type config: dict

        :return: Repository
        """
        return self.repository(MemcachedStore(**config))

    def repository(self, store):
        """
        Create a new cache repository with the given implementation.

        :param store: The cache store implementation instance
        :type store: Store

        :rtype: Repository
        """
        repository = Repository(store)

        return repository

    def _get_prefix(self, config):
        """
        Get the cache prefix.

        :param config: The configuration
        :type config: dict

        :rtype: str
        """
        return config.get('prefix', '')

    def _get_config(self, name):
        """
        Get the cache connection configuration.

        :param name: The cache name
        :type name: str

        :rtype: dict
        """
        return self._config['stores'].get(name)

    def get_default_driver(self):
        """
        Get the default cache driver name.

        :rtype: str

        :raises: RuntimeError
        """
        if 'default' in self._config:
            return self._config['default']

        if len(self._config['stores']) == 1:
            return list(self._config['stores'].keys())[0]

        raise RuntimeError('Missing "default" cache in configuration.')

    def set_default_driver(self, name):
        """
        Set the default cache driver name.

        :param name: The default cache driver name
        :type name: str
        """
        self._config['default'] = name

    def extend(self, driver, store):
        """
        Register a custom driver creator.

        :param driver: The driver
        :type driver: name

        :param store: The store class
        :type store: Store or callable

        :rtype: self
        """
        self._custom_creators[driver] = store

        return self

    def _resolve_serializer(self, serializer):
        """
        Resolve the given serializer.

        :param serializer: The serializer to resolve
        :type serializer: str or Serializer

        :rtype: Serializer
        """
        if isinstance(serializer, Serializer):
            return serializer

        if serializer in self._serializers:
            return self._serializers[serializer]

        raise RuntimeError('Unsupported serializer')

    def register_serializer(self, name, serializer):
        """
        Register a new serializer.

        :param name: The name of the serializer
        :type name: str

        :param serializer: The serializer
        :type serializer: Serializer
        """
        self._serializers[name] = serializer

    def __getattr__(self, item):
        return getattr(self.store(), item)

    def __call__(self, store=None, *args, **kwargs):
        if isinstance(store, (types.FunctionType, types.MethodType)):
            fn = store

            if len(args) > 0:
                store = args[0]
                args = args[1:] if len(args) > 1 else []
            else:
                store = None

            args = (fn,) + args

            return self.store(store)(*args, **kwargs)
        else:
            return self.store(store)(*args, **kwargs)



import os
import re

_escape_pat = re.compile(r'%([0-9A-Fa-f]{2})')
def unescape(v):
    def repl(match):
        n = int(match.group(1), base=16)
        return chr(n)
    return _escape_pat.sub(repl, v)

def parse_addresses(s):
    for addr in s.split(';'):
        transport, info = addr.split(':', 1)
        kv = {}
        for x in info.split(','):
            k, v = x.split('=', 1)
            kv[k] = unescape(v)
        yield (transport, kv)

SUPPORTED_TRANSPORTS = ('unix',)

def get_connectable_addresses(addr):
    unsupported_transports = set()
    found = False
    for transport, kv in parse_addresses(addr):
        if transport not in SUPPORTED_TRANSPORTS:
            unsupported_transports.add(transport)

        elif transport == 'unix':
            if 'abstract' in kv:
                yield '\0' + kv['abstract']
                found = True
            elif 'path' in kv:
                yield kv['path']
                found = True

    if not found:
        raise RuntimeError("DBus transports ({}) not supported. Supported: {}"
                           .format(unsupported_transports, SUPPORTED_TRANSPORTS))

def find_session_bus():
    addr = os.environ['DBUS_SESSION_BUS_ADDRESS']
    return next(get_connectable_addresses(addr))
    # TODO: fallbacks to X, filesystem

def find_system_bus():
    addr = os.environ.get('DBUS_SYSTEM_BUS_ADDRESS', '') \
        or 'unix:path=/var/run/dbus/system_bus_socket'
    return next(get_connectable_addresses(addr))

def get_bus(addr):
    if addr == 'SESSION':
        return find_session_bus()
    elif addr == 'SYSTEM':
        return find_system_bus()
    else:
        return next(get_connectable_addresses(addr))


if __name__ == '__main__':
    print('System bus at:', find_system_bus())
    print('Session bus at:', find_session_bus())



# -*- coding: utf-8 -*-
"""The Apple Framework builds require their own customization"""
import logging
import os
import struct
import subprocess
from abc import ABCMeta, abstractmethod
from textwrap import dedent

from six import add_metaclass

from virtualenv.create.via_global_ref.builtin.ref import ExePathRefToDest, PathRefToDest, RefMust
from virtualenv.info import IS_MAC_ARM64
from virtualenv.util.path import Path
from virtualenv.util.six import ensure_text

from .common import CPython, CPythonPosix, is_mac_os_framework
from .cpython2 import CPython2PosixBase
from .cpython3 import CPython3


@add_metaclass(ABCMeta)
class CPythonmacOsFramework(CPython):
    @classmethod
    def can_describe(cls, interpreter):
        return is_mac_os_framework(interpreter) and super(CPythonmacOsFramework, cls).can_describe(interpreter)

    @classmethod
    def sources(cls, interpreter):
        for src in super(CPythonmacOsFramework, cls).sources(interpreter):
            yield src
        # add a symlink to the host python image
        exe = cls.image_ref(interpreter)
        ref = PathRefToDest(exe, dest=lambda self, _: self.dest / ".Python", must=RefMust.SYMLINK)
        yield ref

    def create(self):
        super(CPythonmacOsFramework, self).create()

        # change the install_name of the copied python executables
        target = "@executable_path/../.Python"
        current = self.current_mach_o_image_path()
        for src in self._sources:
            if isinstance(src, ExePathRefToDest):
                if src.must == RefMust.COPY or not self.symlinks:
                    exes = [self.bin_dir / src.base]
                    if not self.symlinks:
                        exes.extend(self.bin_dir / a for a in src.aliases)
                    for exe in exes:
                        fix_mach_o(str(exe), current, target, self.interpreter.max_size)

    @classmethod
    def _executables(cls, interpreter):
        for _, targets, must, when in super(CPythonmacOsFramework, cls)._executables(interpreter):
            # Make sure we use the embedded interpreter inside the framework, even if sys.executable points to the
            # stub executable in ${sys.prefix}/bin.
            # See http://groups.google.com/group/python-virtualenv/browse_thread/thread/17cab2f85da75951
            fixed_host_exe = Path(interpreter.prefix) / "Resources" / "Python.app" / "Contents" / "MacOS" / "Python"
            yield fixed_host_exe, targets, must, when

    @abstractmethod
    def current_mach_o_image_path(self):
        raise NotImplementedError

    @classmethod
    def image_ref(cls, interpreter):
        raise NotImplementedError


class CPython2macOsFramework(CPythonmacOsFramework, CPython2PosixBase):
    @classmethod
    def image_ref(cls, interpreter):
        return Path(interpreter.prefix) / "Python"

    def current_mach_o_image_path(self):
        return os.path.join(self.interpreter.prefix, "Python")

    @classmethod
    def sources(cls, interpreter):
        for src in super(CPython2macOsFramework, cls).sources(interpreter):
            yield src
        # landmark for exec_prefix
        exec_marker_file, to_path, _ = cls.from_stdlib(cls.mappings(interpreter), "lib-dynload")
        yield PathRefToDest(exec_marker_file, dest=to_path)

    @property
    def reload_code(self):
        result = super(CPython2macOsFramework, self).reload_code
        result = dedent(
            """
        # the bundled site.py always adds the global site package if we're on python framework build, escape this
        import sysconfig
        config = sysconfig.get_config_vars()
        before = config["PYTHONFRAMEWORK"]
        try:
            config["PYTHONFRAMEWORK"] = ""
            {}
        finally:
            config["PYTHONFRAMEWORK"] = before
        """.format(
                result,
            ),
        )
        return result

    @classmethod
    def can_create(cls, interpreter):
        if IS_MAC_ARM64:
            return False
        else:
            return super(CPythonmacOsFramework, cls).can_create(interpreter)


class CPython3macOsFramework(CPythonmacOsFramework, CPython3, CPythonPosix):
    @classmethod
    def image_ref(cls, interpreter):
        return Path(interpreter.prefix) / "Python3"

    def current_mach_o_image_path(self):
        return "@executable_path/../../../../Python3"

    @property
    def reload_code(self):
        result = super(CPython3macOsFramework, self).reload_code
        result = dedent(
            """
        # the bundled site.py always adds the global site package if we're on python framework build, escape this
        import sys
        before = sys._framework
        try:
            sys._framework = None
            {}
        finally:
            sys._framework = before
        """.format(
                result,
            ),
        )
        return result


def fix_mach_o(exe, current, new, max_size):
    """
    https://en.wikipedia.org/wiki/Mach-O

    Mach-O, short for Mach object file format, is a file format for executables, object code, shared libraries,
    dynamically-loaded code, and core dumps. A replacement for the a.out format, Mach-O offers more extensibility and
    faster access to information in the symbol table.

    Each Mach-O file is made up of one Mach-O header, followed by a series of load commands, followed by one or more
    segments, each of which contains between 0 and 255 sections. Mach-O uses the REL relocation format to handle
    references to symbols. When looking up symbols Mach-O uses a two-level namespace that encodes each symbol into an
    'object/symbol name' pair that is then linearly searched for by first the object and then the symbol name.

    The basic structure—a list of variable-length "load commands" that reference pages of data elsewhere in the file—was
    also used in the executable file format for Accent. The Accent file format was in turn, based on an idea from Spice
    Lisp.

    With the introduction of Mac OS X 10.6 platform the Mach-O file underwent a significant modification that causes
    binaries compiled on a computer running 10.6 or later to be (by default) executable only on computers running Mac
    OS X 10.6 or later. The difference stems from load commands that the dynamic linker, in previous Mac OS X versions,
    does not understand. Another significant change to the Mach-O format is the change in how the Link Edit tables
    (found in the __LINKEDIT section) function. In 10.6 these new Link Edit tables are compressed by removing unused and
    unneeded bits of information, however Mac OS X 10.5 and earlier cannot read this new Link Edit table format.
    """
    try:
        logging.debug(u"change Mach-O for %s from %s to %s", ensure_text(exe), current, ensure_text(new))
        _builtin_change_mach_o(max_size)(exe, current, new)
    except Exception as e:
        logging.warning("Could not call _builtin_change_mac_o: %s. " "Trying to call install_name_tool instead.", e)
        try:
            cmd = ["install_name_tool", "-change", current, new, exe]
            subprocess.check_call(cmd)
        except Exception:
            logging.fatal("Could not call install_name_tool -- you must " "have Apple's development tools installed")
            raise


def _builtin_change_mach_o(maxint):
    MH_MAGIC = 0xFEEDFACE
    MH_CIGAM = 0xCEFAEDFE
    MH_MAGIC_64 = 0xFEEDFACF
    MH_CIGAM_64 = 0xCFFAEDFE
    FAT_MAGIC = 0xCAFEBABE
    BIG_ENDIAN = ">"
    LITTLE_ENDIAN = "<"
    LC_LOAD_DYLIB = 0xC

    class FileView(object):
        """A proxy for file-like objects that exposes a given view of a file. Modified from macholib."""

        def __init__(self, file_obj, start=0, size=maxint):
            if isinstance(file_obj, FileView):
                self._file_obj = file_obj._file_obj
            else:
                self._file_obj = file_obj
            self._start = start
            self._end = start + size
            self._pos = 0

        def __repr__(self):
            return "<fileview [{:d}, {:d}] {!r}>".format(self._start, self._end, self._file_obj)

        def tell(self):
            return self._pos

        def _checkwindow(self, seek_to, op):
            if not (self._start <= seek_to <= self._end):
                msg = "{} to offset {:d} is outside window [{:d}, {:d}]".format(op, seek_to, self._start, self._end)
                raise IOError(msg)

        def seek(self, offset, whence=0):
            seek_to = offset
            if whence == os.SEEK_SET:
                seek_to += self._start
            elif whence == os.SEEK_CUR:
                seek_to += self._start + self._pos
            elif whence == os.SEEK_END:
                seek_to += self._end
            else:
                raise IOError("Invalid whence argument to seek: {!r}".format(whence))
            self._checkwindow(seek_to, "seek")
            self._file_obj.seek(seek_to)
            self._pos = seek_to - self._start

        def write(self, content):
            here = self._start + self._pos
            self._checkwindow(here, "write")
            self._checkwindow(here + len(content), "write")
            self._file_obj.seek(here, os.SEEK_SET)
            self._file_obj.write(content)
            self._pos += len(content)

        def read(self, size=maxint):
            assert size >= 0
            here = self._start + self._pos
            self._checkwindow(here, "read")
            size = min(size, self._end - here)
            self._file_obj.seek(here, os.SEEK_SET)
            read_bytes = self._file_obj.read(size)
            self._pos += len(read_bytes)
            return read_bytes

    def read_data(file, endian, num=1):
        """Read a given number of 32-bits unsigned integers from the given file with the given endianness."""
        res = struct.unpack(endian + "L" * num, file.read(num * 4))
        if len(res) == 1:
            return res[0]
        return res

    def mach_o_change(at_path, what, value):
        """Replace a given name (what) in any LC_LOAD_DYLIB command found in the given binary with a new name (value),
        provided it's shorter."""

        def do_macho(file, bits, endian):
            # Read Mach-O header (the magic number is assumed read by the caller)
            cpu_type, cpu_sub_type, file_type, n_commands, size_of_commands, flags = read_data(file, endian, 6)
            # 64-bits header has one more field.
            if bits == 64:
                read_data(file, endian)
            # The header is followed by n commands
            for _ in range(n_commands):
                where = file.tell()
                # Read command header
                cmd, cmd_size = read_data(file, endian, 2)
                if cmd == LC_LOAD_DYLIB:
                    # The first data field in LC_LOAD_DYLIB commands is the offset of the name, starting from the
                    # beginning of the  command.
                    name_offset = read_data(file, endian)
                    file.seek(where + name_offset, os.SEEK_SET)
                    # Read the NUL terminated string
                    load = file.read(cmd_size - name_offset).decode()
                    load = load[: load.index("\0")]
                    # If the string is what is being replaced, overwrite it.
                    if load == what:
                        file.seek(where + name_offset, os.SEEK_SET)
                        file.write(value.encode() + b"\0")
                # Seek to the next command
                file.seek(where + cmd_size, os.SEEK_SET)

        def do_file(file, offset=0, size=maxint):
            file = FileView(file, offset, size)
            # Read magic number
            magic = read_data(file, BIG_ENDIAN)
            if magic == FAT_MAGIC:
                # Fat binaries contain nfat_arch Mach-O binaries
                n_fat_arch = read_data(file, BIG_ENDIAN)
                for _ in range(n_fat_arch):
                    # Read arch header
                    cpu_type, cpu_sub_type, offset, size, align = read_data(file, BIG_ENDIAN, 5)
                    do_file(file, offset, size)
            elif magic == MH_MAGIC:
                do_macho(file, 32, BIG_ENDIAN)
            elif magic == MH_CIGAM:
                do_macho(file, 32, LITTLE_ENDIAN)
            elif magic == MH_MAGIC_64:
                do_macho(file, 64, BIG_ENDIAN)
            elif magic == MH_CIGAM_64:
                do_macho(file, 64, LITTLE_ENDIAN)

        assert len(what) >= len(value)

        with open(at_path, "r+b") as f:
            do_file(f)

    return mach_o_change



from collections import namedtuple
from datetime import datetime
import json
import os
import unittest

from yarg.package import json2package
from yarg.release import Release


class TestRelease(unittest.TestCase):

    def setUp(self):
        package = os.path.join(os.path.dirname(__file__),
                               'package.json')
        self.json = json.loads(open(package).read())
        self.package = json2package(open(package).read())

    def test_release_ids(self):
        self.assertEquals([u'0.0.0', u'0.0.2', u'0.0.15'],
                          self.package.release_ids)

    def test_release(self):
        release_id = '0.0.2'
        release = self.json['releases'][release_id]
        release_list = [Release(release_id, r) for r in release]
        self.assertEquals(release_list[0].md5_digest,
                          self.package.release(release_id)[0].md5_digest)
        self.assertEquals(release_list[1].md5_digest,
                          self.package.release(release_id)[1].md5_digest)
        self.assertEquals('3e3098611177c34706de2e10476b3e50',
                          self.package.release(release_id)[0].md5_digest)
        self.assertEquals('be198baa95116c1c9d17874428e3a0c6',
                          self.package.release(release_id)[1].md5_digest)

    def test_repr(self):
        release_id = '0.0.2'
        release = self.package.release(release_id)[0]
        self.assertEquals(u'<Release 0.0.2>', release.__repr__())

    def test_release_id(self):
        release_id = '0.0.2'
        release = self.package.release(release_id)[0]
        self.assertEquals(release_id, release.release_id)

    def test_release_id(self):
        release_id = '0.0.3'
        release = self.package.release(release_id)
        self.assertEquals(None, release)

    def test_release_uploaded(self):
        release_id = '0.0.2'
        release = self.package.release(release_id)[0]
        self.assertEquals(datetime.strptime("2014-08-16T12:21:20",
                                            "%Y-%m-%dT%H:%M:%S"),
                          release.uploaded)

    def test_release_python_version(self):
        release_id = '0.0.2'
        release = self.package.release(release_id)[0]
        self.assertEquals(u'2.7', release.python_version)

    def test_release_url(self):
        release_id = '0.0.2'
        release = self.package.release(release_id)[0]
        url = u'https://pypi.python.org/packages/2.7/y/yarg/yarg-0.0.2-py2.py3-none-any.whl'
        self.assertEquals(url, release.url)

    def test_release_md5(self):
        release_id = '0.0.2'
        release = self.package.release(release_id)[0]
        md5 = u'3e3098611177c34706de2e10476b3e50'
        self.assertEquals(md5, release.md5_digest)

    def test_release_filename(self):
        release_id = '0.0.2'
        release = self.package.release(release_id)[0]
        filename = u'yarg-0.0.2-py2.py3-none-any.whl'
        self.assertEquals(filename, release.filename)

    def test_release_size(self):
        release_id = '0.0.2'
        release = self.package.release(release_id)[0]
        size = 21596
        self.assertEquals(size, release.size)

    def test_release_unknown_package_type(self):
        release_id = '0.0.0'
        release = self.package.release(release_id)[0]
        self.assertEquals(u'wheeeel', release.package_type)

    def test_release_package_type(self):
        release_id = '0.0.2'
        release = self.package.release(release_id)[0]
        self.assertEquals(u'wheel', release.package_type)

    def test_release_has_sig(self):
        release_id = '0.0.2'
        release = self.package.release(release_id)[0]
        self.assertEquals(True, release.has_sig)

    def test_latest_release_id(self):
        self.assertEquals(u'0.0.15', self.package.latest_release_id)

    def test_latest_release(self):
        release_id = '0.0.15'
        release = self.json['releases'][release_id]
        release_list = [Release(release_id, r) for r in release]
        self.assertEquals(release_list[0].md5_digest,
                          self.package.latest_release[0].md5_digest)
        self.assertEquals(release_list[1].md5_digest,
                          self.package.latest_release[1].md5_digest)
        self.assertEquals('3e3098611177c34706de2e10476b3e51',
                          self.package.latest_release[0].md5_digest)
        self.assertEquals('be198baa95116c1c9d17874428e3a0c7',
                          self.package.latest_release[1].md5_digest)



import os
import sys
from abc import ABC
from errno import ENOENT
from typing import cast

from ._api import BaseFileLock
from ._util import raise_on_exist_ro_file

if sys.platform == "win32":  # pragma: win32 cover
    import msvcrt

    class WindowsFileLock(BaseFileLock):
        """Uses the :func:`msvcrt.locking` function to hard lock the lock file on windows systems."""

        def _acquire(self) -> None:
            raise_on_exist_ro_file(self._lock_file)
            mode = (
                os.O_RDWR  # open for read and write
                | os.O_CREAT  # create file if not exists
                | os.O_TRUNC  # truncate file  if not empty
            )
            try:
                fd = os.open(self._lock_file, mode)
            except OSError as exception:
                if exception.errno == ENOENT:  # No such file or directory
                    raise
            else:
                try:
                    msvcrt.locking(fd, msvcrt.LK_NBLCK, 1)
                except OSError:
                    os.close(fd)
                else:
                    self._lock_file_fd = fd

        def _release(self) -> None:
            fd = cast(int, self._lock_file_fd)
            self._lock_file_fd = None
            msvcrt.locking(fd, msvcrt.LK_UNLCK, 1)
            os.close(fd)

            try:
                os.remove(self._lock_file)
            # Probably another instance of the application hat acquired the file lock.
            except OSError:
                pass


else:  # pragma: win32 no cover

    class WindowsFileLock(BaseFileLock, ABC):
        """Uses the :func:`msvcrt.locking` function to hard lock the lock file on windows systems."""


__all__ = [
    "WindowsFileLock",
]



import sys
import warnings


def saveit(func):
    """A decorator that caches the return value of a function"""

    name = '_' + func.__name__

    def _wrapper(self, *args, **kwds):
        if not hasattr(self, name):
            setattr(self, name, func(self, *args, **kwds))
        return getattr(self, name)
    return _wrapper

cacheit = saveit


def prevent_recursion(default):
    """A decorator that returns the return value of `default` in recursions"""
    def decorator(func):
        name = '_calling_%s_' % func.__name__

        def newfunc(self, *args, **kwds):
            if getattr(self, name, False):
                return default()
            setattr(self, name, True)
            try:
                return func(self, *args, **kwds)
            finally:
                setattr(self, name, False)
        return newfunc
    return decorator


def ignore_exception(exception_class):
    """A decorator that ignores `exception_class` exceptions"""
    def _decorator(func):
        def newfunc(*args, **kwds):
            try:
                return func(*args, **kwds)
            except exception_class:
                pass
        return newfunc
    return _decorator


def deprecated(message=None):
    """A decorator for deprecated functions"""
    def _decorator(func, message=message):
        if message is None:
            message = '%s is deprecated' % func.__name__

        def newfunc(*args, **kwds):
            warnings.warn(message, DeprecationWarning, stacklevel=2)
            return func(*args, **kwds)
        return newfunc
    return _decorator


def cached(size):
    """A caching decorator based on parameter objects"""
    def decorator(func):
        cached_func = _Cached(func, size)
        return lambda *a, **kw: cached_func(*a, **kw)
    return decorator


class _Cached(object):

    def __init__(self, func, count):
        self.func = func
        self.cache = []
        self.count = count

    def __call__(self, *args, **kwds):
        key = (args, kwds)
        for cached_key, cached_result in self.cache:
            if cached_key == key:
                return cached_result
        result = self.func(*args, **kwds)
        self.cache.append((key, result))
        if len(self.cache) > self.count:
            del self.cache[0]
        return result


def resolve(str_or_obj):
    """Returns object from string"""
    from rope.base.utils.pycompat import string_types
    if not isinstance(str_or_obj, string_types):
        return str_or_obj
    if '.' not in str_or_obj:
        str_or_obj += '.'
    mod_name, obj_name = str_or_obj.rsplit('.', 1)
    __import__(mod_name)
    mod = sys.modules[mod_name]
    return getattr(mod, obj_name) if obj_name else mod

def guess_def_lineno(module, node):
    """ Find the line number for a function or class definition.

        `node` may be either an ast.FunctionDef, ast.AsyncFunctionDef, or ast.ClassDef

        Python 3.8 simply provides this to us, but in earlier versions the ast
        node.lineno points to the first decorator rather than the actual
        definition, so we try our best to find where the definitions are.

        This is to workaround bpo-33211 (https://bugs.python.org/issue33211)
    """
    def is_inline_body():
        # class Foo(object):
        #     def inline_body(): pass
        #     ^                  ^--- body_col_offset
        #     `--- indent_col_offset
        #     def not_inline_body():
        #         pass
        #         ^--- body_col_offset == indent_col_offset
        line = module.lines.get_line(node.body[0].lineno)
        indent_col_offset = len(line) - len(line.lstrip())
        body_col_offset = node.body[0].col_offset
        return indent_col_offset < body_col_offset

    if sys.version_info >= (3, 8):
        return node.lineno

    possible_def_line = node.body[0].lineno if is_inline_body() else node.body[0].lineno - 1
    return module.logical_lines.logical_line_in(possible_def_line)[0]



from __future__ import annotations

import importlib.util
import itertools
import os
import re
import shutil
from collections import defaultdict
from collections.abc import Iterator
from typing import IO, TYPE_CHECKING

import pytest
import numpy as np
import numpy.typing as npt
from numpy.typing.mypy_plugin import (
    _PRECISION_DICT,
    _EXTENDED_PRECISION_LIST,
    _C_INTP,
)

try:
    from mypy import api
except ImportError:
    NO_MYPY = True
else:
    NO_MYPY = False

if TYPE_CHECKING:
    # We need this as annotation, but it's located in a private namespace.
    # As a compromise, do *not* import it during runtime
    from _pytest.mark.structures import ParameterSet

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")
PASS_DIR = os.path.join(DATA_DIR, "pass")
FAIL_DIR = os.path.join(DATA_DIR, "fail")
REVEAL_DIR = os.path.join(DATA_DIR, "reveal")
MISC_DIR = os.path.join(DATA_DIR, "misc")
MYPY_INI = os.path.join(DATA_DIR, "mypy.ini")
CACHE_DIR = os.path.join(DATA_DIR, ".mypy_cache")

#: A dictionary with file names as keys and lists of the mypy stdout as values.
#: To-be populated by `run_mypy`.
OUTPUT_MYPY: dict[str, list[str]] = {}


def _key_func(key: str) -> str:
    """Split at the first occurrence of the ``:`` character.

    Windows drive-letters (*e.g.* ``C:``) are ignored herein.
    """
    drive, tail = os.path.splitdrive(key)
    return os.path.join(drive, tail.split(":", 1)[0])


def _strip_filename(msg: str) -> str:
    """Strip the filename from a mypy message."""
    _, tail = os.path.splitdrive(msg)
    return tail.split(":", 1)[-1]


def strip_func(match: re.Match[str]) -> str:
    """`re.sub` helper function for stripping module names."""
    return match.groups()[1]


@pytest.mark.slow
@pytest.mark.skipif(NO_MYPY, reason="Mypy is not installed")
@pytest.fixture(scope="module", autouse=True)
def run_mypy() -> None:
    """Clears the cache and run mypy before running any of the typing tests.

    The mypy results are cached in `OUTPUT_MYPY` for further use.

    The cache refresh can be skipped using

    NUMPY_TYPING_TEST_CLEAR_CACHE=0 pytest numpy/typing/tests
    """
    if (
        os.path.isdir(CACHE_DIR)
        and bool(os.environ.get("NUMPY_TYPING_TEST_CLEAR_CACHE", True))
    ):
        shutil.rmtree(CACHE_DIR)

    for directory in (PASS_DIR, REVEAL_DIR, FAIL_DIR, MISC_DIR):
        # Run mypy
        stdout, stderr, exit_code = api.run([
            "--config-file",
            MYPY_INI,
            "--cache-dir",
            CACHE_DIR,
            directory,
        ])
        if stderr:
            pytest.fail(f"Unexpected mypy standard error\n\n{stderr}")
        elif exit_code not in {0, 1}:
            pytest.fail(f"Unexpected mypy exit code: {exit_code}\n\n{stdout}")
        stdout = stdout.replace('*', '')

        # Parse the output
        iterator = itertools.groupby(stdout.split("\n"), key=_key_func)
        OUTPUT_MYPY.update((k, list(v)) for k, v in iterator if k)


def get_test_cases(directory: str) -> Iterator[ParameterSet]:
    for root, _, files in os.walk(directory):
        for fname in files:
            short_fname, ext = os.path.splitext(fname)
            if ext in (".pyi", ".py"):
                fullpath = os.path.join(root, fname)
                yield pytest.param(fullpath, id=short_fname)


@pytest.mark.slow
@pytest.mark.skipif(NO_MYPY, reason="Mypy is not installed")
@pytest.mark.parametrize("path", get_test_cases(PASS_DIR))
def test_success(path) -> None:
    # Alias `OUTPUT_MYPY` so that it appears in the local namespace
    output_mypy = OUTPUT_MYPY
    if path in output_mypy:
        msg = "Unexpected mypy output\n\n"
        msg += "\n".join(_strip_filename(v) for v in output_mypy[path])
        raise AssertionError(msg)


@pytest.mark.slow
@pytest.mark.skipif(NO_MYPY, reason="Mypy is not installed")
@pytest.mark.parametrize("path", get_test_cases(FAIL_DIR))
def test_fail(path: str) -> None:
    __tracebackhide__ = True

    with open(path) as fin:
        lines = fin.readlines()

    errors = defaultdict(lambda: "")

    output_mypy = OUTPUT_MYPY
    assert path in output_mypy
    for error_line in output_mypy[path]:
        error_line = _strip_filename(error_line)
        match = re.match(
            r"(?P<lineno>\d+): (error|note): .+$",
            error_line,
        )
        if match is None:
            raise ValueError(f"Unexpected error line format: {error_line}")
        lineno = int(match.group('lineno'))
        errors[lineno] += f'{error_line}\n'

    for i, line in enumerate(lines):
        lineno = i + 1
        if (
            line.startswith('#')
            or (" E:" not in line and lineno not in errors)
        ):
            continue

        target_line = lines[lineno - 1]
        if "# E:" in target_line:
            expression, _, marker = target_line.partition("  # E: ")
            expected_error = errors[lineno].strip()
            marker = marker.strip()
            _test_fail(path, expression, marker, expected_error, lineno)
        else:
            pytest.fail(
                f"Unexpected mypy output at line {lineno}\n\n{errors[lineno]}"
            )


_FAIL_MSG1 = """Extra error at line {}

Expression: {}
Extra error: {!r}
"""

_FAIL_MSG2 = """Error mismatch at line {}

Expression: {}
Expected error: {!r}
Observed error: {!r}
"""


def _test_fail(
    path: str,
    expression: str,
    error: str,
    expected_error: None | str,
    lineno: int,
) -> None:
    if expected_error is None:
        raise AssertionError(_FAIL_MSG1.format(lineno, expression, error))
    elif error not in expected_error:
        raise AssertionError(_FAIL_MSG2.format(
            lineno, expression, expected_error, error
        ))


def _construct_ctypes_dict() -> dict[str, str]:
    dct = {
        "ubyte": "c_ubyte",
        "ushort": "c_ushort",
        "uintc": "c_uint",
        "uint": "c_ulong",
        "ulonglong": "c_ulonglong",
        "byte": "c_byte",
        "short": "c_short",
        "intc": "c_int",
        "int_": "c_long",
        "longlong": "c_longlong",
        "single": "c_float",
        "double": "c_double",
        "longdouble": "c_longdouble",
    }

    # Match `ctypes` names to the first ctypes type with a given kind and
    # precision, e.g. {"c_double": "c_double", "c_longdouble": "c_double"}
    # if both types represent 64-bit floats.
    # In this context "first" is defined by the order of `dct`
    ret = {}
    visited: dict[tuple[str, int], str] = {}
    for np_name, ct_name in dct.items():
        np_scalar = getattr(np, np_name)()

        # Find the first `ctypes` type for a given `kind`/`itemsize` combo
        key = (np_scalar.dtype.kind, np_scalar.dtype.itemsize)
        ret[ct_name] = visited.setdefault(key, f"ctypes.{ct_name}")
    return ret


def _construct_format_dict() -> dict[str, str]:
    dct = {k.split(".")[-1]: v.replace("numpy", "numpy.typing") for
           k, v in _PRECISION_DICT.items()}

    return {
        "uint8": "numpy.unsignedinteger[numpy.typing._8Bit]",
        "uint16": "numpy.unsignedinteger[numpy.typing._16Bit]",
        "uint32": "numpy.unsignedinteger[numpy.typing._32Bit]",
        "uint64": "numpy.unsignedinteger[numpy.typing._64Bit]",
        "uint128": "numpy.unsignedinteger[numpy.typing._128Bit]",
        "uint256": "numpy.unsignedinteger[numpy.typing._256Bit]",
        "int8": "numpy.signedinteger[numpy.typing._8Bit]",
        "int16": "numpy.signedinteger[numpy.typing._16Bit]",
        "int32": "numpy.signedinteger[numpy.typing._32Bit]",
        "int64": "numpy.signedinteger[numpy.typing._64Bit]",
        "int128": "numpy.signedinteger[numpy.typing._128Bit]",
        "int256": "numpy.signedinteger[numpy.typing._256Bit]",
        "float16": "numpy.floating[numpy.typing._16Bit]",
        "float32": "numpy.floating[numpy.typing._32Bit]",
        "float64": "numpy.floating[numpy.typing._64Bit]",
        "float80": "numpy.floating[numpy.typing._80Bit]",
        "float96": "numpy.floating[numpy.typing._96Bit]",
        "float128": "numpy.floating[numpy.typing._128Bit]",
        "float256": "numpy.floating[numpy.typing._256Bit]",
        "complex64": ("numpy.complexfloating"
                      "[numpy.typing._32Bit, numpy.typing._32Bit]"),
        "complex128": ("numpy.complexfloating"
                       "[numpy.typing._64Bit, numpy.typing._64Bit]"),
        "complex160": ("numpy.complexfloating"
                       "[numpy.typing._80Bit, numpy.typing._80Bit]"),
        "complex192": ("numpy.complexfloating"
                       "[numpy.typing._96Bit, numpy.typing._96Bit]"),
        "complex256": ("numpy.complexfloating"
                       "[numpy.typing._128Bit, numpy.typing._128Bit]"),
        "complex512": ("numpy.complexfloating"
                       "[numpy.typing._256Bit, numpy.typing._256Bit]"),

        "ubyte": f"numpy.unsignedinteger[{dct['_NBitByte']}]",
        "ushort": f"numpy.unsignedinteger[{dct['_NBitShort']}]",
        "uintc": f"numpy.unsignedinteger[{dct['_NBitIntC']}]",
        "uintp": f"numpy.unsignedinteger[{dct['_NBitIntP']}]",
        "uint": f"numpy.unsignedinteger[{dct['_NBitInt']}]",
        "ulonglong": f"numpy.unsignedinteger[{dct['_NBitLongLong']}]",
        "byte": f"numpy.signedinteger[{dct['_NBitByte']}]",
        "short": f"numpy.signedinteger[{dct['_NBitShort']}]",
        "intc": f"numpy.signedinteger[{dct['_NBitIntC']}]",
        "intp": f"numpy.signedinteger[{dct['_NBitIntP']}]",
        "int_": f"numpy.signedinteger[{dct['_NBitInt']}]",
        "longlong": f"numpy.signedinteger[{dct['_NBitLongLong']}]",

        "half": f"numpy.floating[{dct['_NBitHalf']}]",
        "single": f"numpy.floating[{dct['_NBitSingle']}]",
        "double": f"numpy.floating[{dct['_NBitDouble']}]",
        "longdouble": f"numpy.floating[{dct['_NBitLongDouble']}]",
        "csingle": ("numpy.complexfloating"
                    f"[{dct['_NBitSingle']}, {dct['_NBitSingle']}]"),
        "cdouble": ("numpy.complexfloating"
                    f"[{dct['_NBitDouble']}, {dct['_NBitDouble']}]"),
        "clongdouble": (
            "numpy.complexfloating"
            f"[{dct['_NBitLongDouble']}, {dct['_NBitLongDouble']}]"
        ),

        # numpy.typing
        "_NBitInt": dct['_NBitInt'],

        # numpy.ctypeslib
        "c_intp": f"ctypes.{_C_INTP}"
    }


#: A dictionary with all supported format keys (as keys)
#: and matching values
FORMAT_DICT: dict[str, str] = _construct_format_dict()
FORMAT_DICT.update(_construct_ctypes_dict())


def _parse_reveals(file: IO[str]) -> tuple[npt.NDArray[np.str_], list[str]]:
    """Extract and parse all ``"  # E: "`` comments from the passed
    file-like object.

    All format keys will be substituted for their respective value
    from `FORMAT_DICT`, *e.g.* ``"{float64}"`` becomes
    ``"numpy.floating[numpy.typing._64Bit]"``.
    """
    string = file.read().replace("*", "")

    # Grab all `# E:`-based comments and matching expressions
    expression_array, _, comments_array = np.char.partition(
        string.split("\n"), sep="  # E: "
    ).T
    comments = "/n".join(comments_array)

    # Only search for the `{*}` pattern within comments, otherwise
    # there is the risk of accidentally grabbing dictionaries and sets
    key_set = set(re.findall(r"\{(.*?)\}", comments))
    kwargs = {
        k: FORMAT_DICT.get(k, f"<UNRECOGNIZED FORMAT KEY {k!r}>") for
        k in key_set
    }
    fmt_str = comments.format(**kwargs)

    return expression_array, fmt_str.split("/n")


@pytest.mark.slow
@pytest.mark.skipif(NO_MYPY, reason="Mypy is not installed")
@pytest.mark.parametrize("path", get_test_cases(REVEAL_DIR))
def test_reveal(path: str) -> None:
    """Validate that mypy correctly infers the return-types of
    the expressions in `path`.
    """
    __tracebackhide__ = True

    with open(path) as fin:
        expression_array, reveal_list = _parse_reveals(fin)

    output_mypy = OUTPUT_MYPY
    assert path in output_mypy
    for error_line in output_mypy[path]:
        error_line = _strip_filename(error_line)
        match = re.match(
            r"(?P<lineno>\d+): note: .+$",
            error_line,
        )
        if match is None:
            raise ValueError(f"Unexpected reveal line format: {error_line}")
        lineno = int(match.group('lineno')) - 1
        assert "Revealed type is" in error_line

        marker = reveal_list[lineno]
        expression = expression_array[lineno]
        _test_reveal(path, expression, marker, error_line, 1 + lineno)


_REVEAL_MSG = """Reveal mismatch at line {}

Expression: {}
Expected reveal: {!r}
Observed reveal: {!r}
"""


def _test_reveal(
    path: str,
    expression: str,
    reveal: str,
    expected_reveal: str,
    lineno: int,
) -> None:
    """Error-reporting helper function for `test_reveal`."""
    strip_pattern = re.compile(r"(\w+\.)+(\w+)")
    stripped_reveal = strip_pattern.sub(strip_func, reveal)
    stripped_expected_reveal = strip_pattern.sub(strip_func, expected_reveal)
    if stripped_reveal not in stripped_expected_reveal:
        raise AssertionError(
            _REVEAL_MSG.format(lineno,
                               expression,
                               stripped_expected_reveal,
                               stripped_reveal)
        )


@pytest.mark.slow
@pytest.mark.skipif(NO_MYPY, reason="Mypy is not installed")
@pytest.mark.parametrize("path", get_test_cases(PASS_DIR))
def test_code_runs(path: str) -> None:
    """Validate that the code in `path` properly during runtime."""
    path_without_extension, _ = os.path.splitext(path)
    dirname, filename = path.split(os.sep)[-2:]

    spec = importlib.util.spec_from_file_location(
        f"{dirname}.{filename}", path
    )
    assert spec is not None
    assert spec.loader is not None

    test_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(test_module)


LINENO_MAPPING = {
    3: "uint128",
    4: "uint256",
    6: "int128",
    7: "int256",
    9: "float80",
    10: "float96",
    11: "float128",
    12: "float256",
    14: "complex160",
    15: "complex192",
    16: "complex256",
    17: "complex512",
}


@pytest.mark.slow
@pytest.mark.skipif(NO_MYPY, reason="Mypy is not installed")
def test_extended_precision() -> None:
    path = os.path.join(MISC_DIR, "extended_precision.pyi")
    output_mypy = OUTPUT_MYPY
    assert path in output_mypy

    with open(path, "r") as f:
        expression_list = f.readlines()

    for _msg in output_mypy[path]:
        *_, _lineno, msg_typ, msg = _msg.split(":")

        msg = _strip_filename(msg)
        lineno = int(_lineno)
        expression = expression_list[lineno - 1].rstrip("\n")
        msg_typ = msg_typ.strip()
        assert msg_typ in {"error", "note"}

        if LINENO_MAPPING[lineno] in _EXTENDED_PRECISION_LIST:
            if msg_typ == "error":
                raise ValueError(f"Unexpected reveal line format: {lineno}")
            else:
                marker = FORMAT_DICT[LINENO_MAPPING[lineno]]
                _test_reveal(path, expression, marker, msg, lineno)
        else:
            if msg_typ == "error":
                marker = "Module has no attribute"
                _test_fail(path, expression, marker, msg, lineno)



from typing import List

from .solution import Solution


class BaseSolution(Solution):
    def __init__(self, title: str = None, description: str = None) -> None:
        self._title = title
        self._description = description
        self._links = []

    @property
    def solution_title(self) -> str:
        return self._title

    @property
    def solution_description(self) -> str:
        return self._description

    @property
    def documentation_links(self) -> List[str]:
        return self._links
